/****************************************************************************

  $Source: /home/foda/fracas/RCS/computers.c $
  $Revision: 7.12 $
  $Date: 1998/12/17 16:50:28 $
  $State: Exp $
  $Locker:  $

****************************************************************************/
#ident "$Id: computers.c 7.12 1998/12/17 16:50:28 pot Exp $"

/*
 *  FRACAS stands for FRAmed Channel Access Simulator.
 */

#include "fracas.h"
#include "libnr.h"

#define DEFAULT_COLS 78
#define DEFAULT_ROWS 22

static computer_r simplestats_reader;
static computer_r listing_reader;
static computer_r quantile_reader;
static computer_r categorise_reader;
static computer_r ttydistrib_reader;
static computer_r ttygraph_reader;

computer_n computers [] =
{
  { simplestats_reader, "simplestats" },
  { listing_reader, "listing" },
  { quantile_reader, "quantile" },
  { categorise_reader, "categorise" },
  { ttydistrib_reader, "ttydistrib" },
  { ttygraph_reader, "ttygraph" },
  { NULL }
};

#define print_headers (outstream == stdout && !terse)

/*
 * Return the observable from the argument string.
 */
static observable
extract_observable (char *lp)
{
  observable_name *on;

  if (lp == NULL)
    {
      strcpy (errmsg, "missing observable");
      return (wrong_o);
    }
  for (on = observable_names; on->what != LAST; on++)
    if (tokenis (on->name))
      return (on->what);
  strcpy (errmsg, "unknown observable");
  return (wrong_o);
} /* extract_observable() */

/*
 * For each station from FIRSTNO to LASTNO allocate a history structure
 * for WHAT and add it at the head of the station history list.
 * Station number 0 is the sum of the observables of the other stations.
 */
static void
add_history (int firstno, int lastno, observable what)
{
  int j;
  unsigned long hlen = histlen;

  switch (what)
    {
    case wrong_o:
      internal_error ();
    case s_trudelay_o:
      hlen = trudhistlen;
    case s_delay_o:
      if (firstno == 0) lastno = stno;
      add_history ((firstno==0 ? 1 : firstno), lastno, s_input_o);
      add_history ((firstno==0 ? 1 : firstno), lastno, s_sent_o);
      break;
    case v_trudelay_o:
      hlen = trudhistlen;
    case v_delay_o:
      if (firstno == 0) lastno = stno;
      add_history ((firstno==0 ? 1 : firstno), lastno, v_input_o);
      add_history ((firstno==0 ? 1 : firstno), lastno, v_sent_o);
      break;
    case i_trudelay_o:
      hlen = trudhistlen;
    case i_delay_o:
      if (firstno == 0) lastno = stno;
      add_history ((firstno==0 ? 1 : firstno), lastno, i_input_o);
      add_history ((firstno==0 ? 1 : firstno), lastno, i_sent_o);
      break;
    case b_trudelay_o:
      hlen = trudhistlen;
    case b_delay_o:
      if (firstno == 0) lastno = stno;
      add_history ((firstno==0 ? 1 : firstno), lastno, b_input_o);
      add_history ((firstno==0 ? 1 : firstno), lastno, b_sent_o);
      break;
    default:
      break;
    }

  for (j = firstno; j <= lastno; j++)
    {
      int hl;
      history *hp;
      station *stp = &st_table[j];

      for (hp = stp->history_list; hp != NULL; hp = hp->h_next)
	if (hp->h_what == what)
	  goto next_station;

      hl = hlen;
      if (j == 0 && (what == s_trudelay_o
		     || what == v_trudelay_o
		     || what == i_trudelay_o
		     || what == b_trudelay_o))
	hl *= stno;
      hp = allocate (sizeof (history) + (hl-1) * sizeof (history_item));
      hp->h_what = what;
      hp->h_last = &hp->h_items[hl-1];
      hp->h_current = hp->h_items;
      hp->h_next = stp->history_list;
      stp->history_list = hp;

    next_station:
      continue;
    }
} /* add_history() */



/*
 * For each station from FIRSTNO to LASTNO allocate a collection
 * structure for WHAT and add it at the head of the station
 * collection list.
 * Station number 0 is the sum of the observables of the other stations.
 */
static void
add_collection (int firstno, int lastno, observable what)
{
  int j;

  switch (what)
    {
    case wrong_o:
      internal_error ();
    case s_delay_o:
      if (firstno == 0) add_history (1, stno, what);
    case s_trudelay_o:
      if (firstno == 0) lastno = stno;
      add_history ((firstno==0 ? 1 : firstno), lastno, s_input_o);
      add_history ((firstno==0 ? 1 : firstno), lastno, s_sent_o);
      break;
    case v_delay_o:
      if (firstno == 0) add_history (1, stno, what);
    case v_trudelay_o:
      if (firstno == 0) lastno = stno;
      add_history ((firstno==0 ? 1 : firstno), lastno, v_input_o);
      add_history ((firstno==0 ? 1 : firstno), lastno, v_sent_o);
      break;
    case i_delay_o:
      if (firstno == 0) add_history (1, stno, what);
    case i_trudelay_o:
      if (firstno == 0) lastno = stno;
      add_history ((firstno==0 ? 1 : firstno), lastno, i_input_o);
      add_history ((firstno==0 ? 1 : firstno), lastno, i_sent_o);
      break;
    case b_delay_o:
      if (firstno == 0) add_history (1, stno, what);
    case b_trudelay_o:
      if (firstno == 0) lastno = stno;
      add_history ((firstno==0 ? 1 : firstno), lastno, b_input_o);
      add_history ((firstno==0 ? 1 : firstno), lastno, b_sent_o);
      break;
    default:
      break;
    }

  for (j = firstno; j <= lastno; j++)
    {
      collection *cp;
      station *stp = &st_table[j];

      for (cp = stp->collection_list; cp != NULL; cp = cp->c_next)
	if (cp->c_what == what)
	  goto next_station;

      cp = allocate (sizeof (collection));
      cp->c_what = what;
      st_init (&cp->c_data);
      cp->c_next = stp->collection_list;
      stp->collection_list = cp;

    next_station:
      continue;
    }
} /* add_collection () */



/*
 * Allocate a computation structure and add it to the list whose head
 * is pointed by LISTP.
 */
static computation *
append_computation (computer *comp, int stndx, observable what,
		    computation **listp)
{
  computation *cp;

  cp = allocate (sizeof (computation));
  cp->cm_st = stndx;
  cp->cm_what = what;
  cp->cm_computer = comp;
  cp->cm_next = NULL;

  while (*listp != NULL)
    listp = &(*listp)->cm_next;
  *listp = cp;

  return (cp);
} /* append_computation() */

typedef struct
{
  history_item *a, *b;
  st_data *c;
}
sum_data;

/* Flags and variables for the warnings of truncated delay (max delay
   is limited by the size of the `dshort' field `n') and of insufficient
   trudelay histroy length and consequent loss of some delays. */
static logical trud_big_warning;
static int trud_big_no;
static unsigned int trud_big_max;
static logical histlen_short_warning;
static int histlen_short_no;

/*
 * Given a VAL and an N, put them in the dshort structure pointed by
 * DSP.  Make overflow checks, because if N is greater than the
 * max number that fits in dsp->f, more than one dshort is occupied.
 * LAST_DSP is used in order not to overflow the dshort array.
 */
static int
add_ds_to_history (dshort *dsp, dshort *last_dsp,
		   unsigned int val, unsigned int n)
{
  int incr = 0;

  if (val > 2*SHRT_MAX+1)
    {
      if (!trud_big_warning)
	{
	  runtime_error ("some tru_delay's are too big and will be truncated");
	  trud_big_no = 1;
	  trud_big_max = val;
	  trud_big_warning = TRUE;
	}
      else
	{
	  trud_big_no += 1;
	  if (val > trud_big_max)
	    trud_big_max = val;
	}
      val = 2*SHRT_MAX+1;
    }

  while (n > 2*SHRT_MAX+1 && dsp <= last_dsp)
    {
      dsp->n = 2*SHRT_MAX+1;
      dsp->val = val;
      dsp += 1;
      incr += 1;
      n -= 2*SHRT_MAX+1;
    }
  if (dsp <= last_dsp)
    {
      dsp->n = n;
      dsp->val = val;
      incr += 1;
    }

  if (incr == 0)
    {
      if (!histlen_short_warning)
	{
	  runtime_error ("trudhistlen too short: some tru_delay will be lost");
	  histlen_short_no = 1;
	  histlen_short_warning = TRUE;
	}
      else
	histlen_short_no += 1;
    }

  return (incr);
} /* add_ds_to_history() */

/*
   COMPUTING THE DELAY
   A discussion on the error of the delay estimation in Fracas

   In Fracas every packet is generated in the current frame, without
   any knowledge of the instant in the frame when it is generated.
   Nor is there any knowledge about the instant in the frame when the
   packet is sent.  Fracas's idea of delay is that, if a packet is
   generated in a frame (the packet arrives at the station) and there
   is an available allocation after N frames (the packet leaves the
   station), the packet has a delay of N frames. In a real system, on
   the opposite, the packet's delay would lay in the range [N-1; N+1]
   frames, depending on the position in the frame when the packet
   arrives and the position in the frame when it leaves.

   We have then an error in the range [-1; 1] frames, which is the sum
   of two errors: the position of the arriving packet in the frame and
   the position of the leaving packet in the frame, each in the range
   [-0.5; 0.5] frames.

   The second error is not easily estimated, because it depends on the
   position of the allocation inside the frame, which can have any
   distribution (and thus any mean) inside the range.

   The first error can be estimated with high accuracy if we suppose
   that the instants when the packets arrive have no correlation with
   the frame repetition period.  With this assumption, the error for
   each packet has a uniform distribution in the error range.  If we
   assume an emulation run with a number of frames k (say, k >> 100),
   the error of the mean has a gaussian distribution with null mean
   and variance equal to 1/12k frames, which is negligible for any
   practical purpose.

   We can then assume that the error of the mean delay computed by
   fracas has an error in the range [-0.5; 0.5] frames.
*/

/*
 * Compute the delay of the TRUs entering a station.
 * IHP and SHP are the history pointers of the input and sent TRUs.
 * QUEUE is the number of TRUs in the queue at the beginning of time.
 * DHP and DCP, if not NULL, point to the delay history and collection
 * to be filled with the values computed.
 */
static void
station_tru_delay (history *ihp, history *shp, int queue,
		   history *dhp, collection *dcp)
{
  register int output;
  register int input;
  register int group;
  register int frdelay;
  register float *ip;
  register float *sp;
  register float *past_last_sp;
  register float *past_last_ip;

  const int frtime = frametime;
  const int rtdelay = rttime;	/* see above: "COMPUTING THE DELAY" */

  ip = &ihp->h_items->f;
  sp = &shp->h_items->f;
  past_last_sp = &shp->h_current->f;
  past_last_ip = &ihp->h_current->f;
  output = -queue;
  frdelay = rtdelay;

  while (sp < past_last_sp && ip < past_last_ip)
    {
      /* Here input is null or uninitialised. */
      input = *ip++;
      frdelay -= frtime;

      while (input > 0)
	{
	  while (output <= 0)
	    {
	      if (sp == past_last_sp)
		return;
	      output += *sp++;
	      frdelay += frtime;
	    }

	  group = MIN (output, input);
	  output -= group;
	  input -= group;

	  if (dhp != NULL)
	    dhp->h_current +=
	      add_ds_to_history (&dhp->h_current->ds, &dhp->h_last->ds,
				 frdelay, group);
	  if (dcp != NULL)
	    st_collect (&dcp->c_data, frdelay, group);
	} /* while (input > 0) */
    } /* while the sent pointer sp scans the whole sent history */

  if (trud_big_warning)
    {
      char errbuf [BUFSIZ];
      sprintf (errbuf, "%d tru_delay's (highest was %u) truncated to %u",
	       trud_big_no, trud_big_max, 2*SHRT_MAX+1);
      runtime_error (errbuf);
      trud_big_warning = FALSE;
    }
  if (histlen_short_warning)
    {
      char errbuf [BUFSIZ];
      sprintf (errbuf, "space for %d items only: %d tru_delay lost",
	       dhp->h_current - dhp->h_items, histlen_short_no);
      runtime_error (errbuf);
      histlen_short_warning = FALSE;
    }
} /* station_tru_delay() */

/*
 * Compute the delay of the TRUs entering the whole system, that is,
 * copy the delays of all the stations in a single long history.  The
 * result makes no sense if viewed as a function of time: quantile can
 * be computed, but neither correlation nor time graph are meaningful.
 * DHP and DCP, if not NULL, point to the delay history and collection
 * to be filled with the values computed.
 * FLDATA[stno+1] is an array of sum_data elements: FLDATA[j].a
 * points to the first, FLDATA[j].b points to the last history item
 * of station j, FLDATA[j].c points to the collection data of station j.
 */
static void
sum_tru_delay (sum_data fldata[],
	       history *dhp, collection *dcp)
{
  int j;

  if (dhp != NULL)
    for (j = 1; j <= stno; j++)
      {
	history_item *itemp = fldata[j].a;
	history_item *lastp = fldata[j].b;

	while (itemp < lastp)
	  *dhp->h_current++ = *itemp++;
      }

  if (dcp != NULL)
    {
      dcp->c_data = *fldata[1].c;
      for (j = 2; j <= stno; j++)
	st_add (&dcp->c_data, fldata[j].c);
    }

} /* sum_tru_delay() */

/*
 * Compute the delay of the TRUs entering a station in each frame.
 * IHP and SHP are the history pointers of the input and sent TRUs.
 * QUEUE is the number of TRUs in the queue at the beginning of time.
 * DHP and DCP, if not NULL, point to the delay history and collection
 * to be filled with the values computed.
 */
static void
station_frame_delay (history *ihp, history *shp, int queue,
		     history *dhp, collection *dcp)
{
  register int frameinput;
  register int output;
  register int group;
  register int input;
  register int delay;
  register int frdelay;
  float result;
  register float *ip;
  register float *sp;
  register float *past_last_sp;
  register float *past_last_ip;

  const int frtime = frametime;
  const int rtdelay = rttime;	/* see above: "COMPUTING THE DELAY" */

  ip = &ihp->h_items->f;
  sp = &shp->h_items->f;
  past_last_sp = &shp->h_current->f;
  past_last_ip = &ihp->h_current->f;
  output = -queue;
  frdelay = rtdelay;
  while (sp < past_last_sp && ip < past_last_ip)
    {
      /* Here input is null or uninitialised. */
      input = frameinput = *ip++;
      delay = 0;
      frdelay -= frtime;

      while (input > 0)
	{
	  while (output <= 0)
	    {
	      if (sp == past_last_sp)
		return;
	      output += *sp++;
	      frdelay += frtime;
	    }

	  group = MIN (output, input);
	  delay += group * frdelay;
	  output -= group;
	  input -= group;
	}

      if (frameinput == 0)
	{
	  /* A history value of 0 conventionally means that no packets
	     have been sent in that frame. */
	  if (dhp != NULL)
	    (dhp->h_current++)->f = 0;
	}
      else /* if any input packets in this frame */
	{
	  result = (float)delay / frameinput;
	  if (dhp != NULL)
	    (dhp->h_current++)->f = result;
	  if (dcp != NULL)
	    st_collect (&dcp->c_data, result, 1);
	}

    } /* for the whole input history */
} /* station_frame_delay() */

/*
 * Compute the delay of the TRUs entering the whole system in each frame.
 * DHP and DCP, if not NULL, point to the delay history and collection
 * to be filled with the values computed.
 * LEN is the length of the longest frame delay history of all the stations.
 * SDDATA[stno+1] is an array of sum_data elements: SDDATA[j].a
 * points to the first sent, SDDATA[j].b points to the first delay
 * history item of station j.
 */
static void
sum_frame_delay (sum_data sddata[], int len,
		 history *dhp, collection *dcp)
{
  for (; len > 0; len--)
    {
      int j;
      float avg_delay;
      float delay_sum = 0;
      float sent_sum = 0;

      for (j = 1; j <= stno; j++)
	{
	  float sent = (sddata[j].a++)->f;
	  float delay = (sddata[j].b++)->f;

	  delay_sum += delay * sent;
	  sent_sum += sent;
	}
      avg_delay = delay_sum / sent_sum;

      if (dhp != NULL)
	(dhp->h_current++)->f = avg_delay;
      if (dcp != NULL)
	st_collect (&dcp->c_data, avg_delay, 1);
    } /* for the maximum frame delay history lenght of all stations */
} /* sum_frame_delay() */

/*
 * Call the correct station delay computer, choosing among
 * station_tru_delay and station_frame_delay.
 * The choice is made on the basis of DELAYOBS.  DHP and DCP are the
 * relevant delay history and collection pointer, if not NULL.
 */
static void
call_station_delay_computer (station *stp, observable delayobs,
			     history *dhp, collection *dcp)
{
  history *ihp, *shp;
  observable inputobs, sentobs;
  int queue;
  logical framedelay;

  framedelay = FALSE;
  switch (delayobs)
    {
    case s_delay_o:
      framedelay = TRUE;
    case s_trudelay_o:
      if (stp->trus_dropped.s) goto dropped_packets;
      queue = stp->warmup_queue.s;
      inputobs = s_input_o; sentobs = s_sent_o;
      break;
    case v_delay_o:
      framedelay = TRUE;
    case v_trudelay_o:
      if (stp->trus_dropped.v) goto dropped_packets;
      queue = stp->warmup_queue.v;
      inputobs = v_input_o; sentobs = v_sent_o;
      break;
    case i_delay_o:
      framedelay = TRUE;
    case i_trudelay_o:
      if (stp->trus_dropped.i) goto dropped_packets;
      queue = stp->warmup_queue.i;
      inputobs = i_input_o; sentobs = i_sent_o;
      break;
    case b_delay_o:
      framedelay = TRUE;
    case b_trudelay_o:
      if (stp->trus_dropped.b) goto dropped_packets;
      queue = stp->warmup_queue.b;
      inputobs = b_input_o; sentobs = b_sent_o;
      break;
    default:
      queue = 0; inputobs = sentobs = wrong_o; internal_error ();
    }

  for (ihp = stp->history_list; ihp != NULL; ihp = ihp->h_next)
    if (ihp->h_what == inputobs)
      break;
  for (shp = stp->history_list; shp != NULL; shp = shp->h_next)
    if (shp->h_what == sentobs)
      break;
  ASSERT (ihp != NULL && shp != NULL);

  if (framedelay)
    station_frame_delay (ihp, shp, queue, dhp, dcp);
  else
    station_tru_delay (ihp, shp, queue, dhp, dcp);
  return;

  if (trud_big_warning)
    {
    }
  if (histlen_short_warning)
    {
    }

  dropped_packets:
  {
    char errbuf [BUFSIZ];
    sprintf (errbuf, "TRUs dropped prevent computing of `%s' for station %ld",
	     observable_string (delayobs), stp->number);
    runtime_error (errbuf);
    return;
  }
} /* call_station_delay_computer() */

/*
 * Call the correct sum delay computer, choosing among sum_tru_delay
 * and sum_frame_delay.
 * The choice is made on the basis of DELAYOBS.  DHP and DCP are the
 * relevant delay history and collection pointer, if not NULL.
 */
static void
call_sum_delay_computer (observable delayobs,
			 history *dhp, collection *dcp)
{
  int j;
  observable inputobs, sentobs;
  logical framedelay;
  int maxhlen = 0;
  sum_data *data = allocate ((stno+1) * sizeof (sum_data));

  framedelay = FALSE;
  switch (delayobs)
    {
    case s_delay_o:
      framedelay = TRUE;
    case s_trudelay_o:
      inputobs = s_input_o; sentobs = s_sent_o;
      break;
    case v_delay_o:
      framedelay = TRUE;
    case v_trudelay_o:
      inputobs = v_input_o; sentobs = v_sent_o;
      break;
    case i_delay_o:
      framedelay = TRUE;
    case i_trudelay_o:
      inputobs = i_input_o; sentobs = i_sent_o;
      break;
    case b_delay_o:
      framedelay = TRUE;
    case b_trudelay_o:
      inputobs = b_input_o; sentobs = b_sent_o;
      break;
    default:
      inputobs = sentobs = wrong_o; internal_error ();
    }

  for (j = 1; j <= stno; j++)
    {
      history *sihp, *sdhp;
      collection *sdcp;
      station *stp = &st_table[j];

      for (sihp = stp->history_list; sihp != NULL; sihp = sihp->h_next)
	if (sihp->h_what == inputobs)
	  break;
      for (sdhp = stp->history_list; sdhp != NULL; sdhp = sdhp->h_next)
	if (sdhp->h_what == delayobs)
	  break;
      for (sdcp = stp->collection_list; sdcp != NULL; sdcp = sdcp->c_next)
	if (sdcp->c_what == delayobs)
	  break;
      if (sihp == NULL
	  || (framedelay && sdhp == NULL)
	  || (!framedelay && ((dhp != NULL && sdhp == NULL)
			      || (dcp != NULL && sdcp == NULL))))
	internal_error ();
      if (framedelay)
	{
	  int hlen = sdhp->h_current - sdhp->h_items;
	  if (hlen > maxhlen)
	    maxhlen = hlen;	/* max length of the delay histories */
	  data[j].a = sihp->h_items;	/* first input item */
	  data[j].b = sdhp->h_items;	/* first delay item */
	}
      else /* TRU delay */
	{
	  if (dhp != NULL)
	    {
	      data[j].a = sdhp->h_items; /* first delay item */
	      data[j].b = sdhp->h_current; /* last delay item */
	    }
	  if (dcp != NULL)
	    data[j].c = &sdcp->c_data; /* collection data */
	}
    } /* for all the stations */

  if (framedelay)
    sum_frame_delay (data, maxhlen, dhp, dcp);
  else
    sum_tru_delay (data, dhp, dcp);

  free (data);
} /* call_sum_delay_computer() */

void
compute_delay_sum (void)
{
  compute_station_delay (NULL);
}

/*
 * This function fills the delay histories (if present) with the
 * proper values.  When a delay history is present, it expects to find
 * also an input history and a sent history of the same type of
 * traffic.  If a delay collection of the same type is also present,
 * it is filled with the proper values.
 * If STP is NULL, the sum of the delays is computed.
 */
void
compute_station_delay (station *stp)
{
  history *dhp;
  collection *dcp;
  logical compute_sum;

  if (stp == NULL)
    {
      compute_sum = TRUE;
      stp = &st_table[0];
    }
  else
    compute_sum = FALSE;

  /* Look for a delay history. */
  for (dhp = stp->history_list; dhp != NULL; dhp = dhp->h_next)
    {
      switch (dhp->h_what)
	{
	case s_delay_o: case s_trudelay_o:
	case v_delay_o: case v_trudelay_o:
	case i_delay_o: case i_trudelay_o:
	case b_delay_o: case b_trudelay_o:
	  for (dcp = stp->collection_list; dcp != NULL; dcp = dcp->c_next)
	    if (dcp->c_what == dhp->h_what)
	      break;
	  if (compute_sum)
	    call_sum_delay_computer (dhp->h_what, dhp, dcp);
	  else
	    call_station_delay_computer (stp, dhp->h_what, dhp, dcp);
	default:
	  ;
	}
    } /* for all the delay histories in st */

  /* Look for the yet uncomputed delay collections. */
  for (dcp = stp->collection_list; dcp != NULL; dcp = dcp->c_next)
    {
      switch (dcp->c_what)
	{
	case s_delay_o: case s_trudelay_o:
	case v_delay_o: case v_trudelay_o:
	case i_delay_o: case i_trudelay_o:
	case b_delay_o: case b_trudelay_o:
	  if (st_n (&dcp->c_data) > 0)
	    continue;
	  if (compute_sum)
	    call_sum_delay_computer (dcp->c_what, NULL, dcp);
	  else
	    call_station_delay_computer (stp, dcp->c_what, NULL, dcp);
	default:
	  ;
	}
    } /* for all the delay histories in st */
} /* compute_station_delay() */

/*
 * Print simple statistics of observable data.
 *
 * Syntax is: simplestats <observable>
 *	number of samples, minimum, maximum, mean, variance and standard
 *	deviation are printed.
 */
static void
simplestats_computer (station *stp, observable what, cm_args *argsp)
{
  collection *cp;
  FILE *outstream = argsp->simplestats.ostream;

  for (cp = stp->collection_list; cp != NULL; cp = cp->c_next)
    if (cp->c_what == what)
      {
	double samples = st_n (&cp->c_data);
        if (print_headers)
	  {
	    printf ("Number of `%s' samples for station #%ld is %.0f.",
		    observable_string (what), stp->number, samples);
	    printf ("  Statistics:\n");
	    printf ("%15s\t%s\t%s", "samples", "minimum", "maximum");
	    if (samples >= 2)
	      printf ("\t%s\t%s\t%s", "mean", "var", "stddev");
	    printf ("\n");
	  }
	fprintf (outstream, "%15.0f\t%g\t%g",
		 st_n (&cp->c_data),
		 st_min (&cp->c_data), st_max (&cp->c_data));
	if (samples >= 2)
	  fprintf (outstream,  "\t%g\t%g\t%g\n",
		   st_avg (&cp->c_data),
		   st_var (&cp->c_data), st_stddev (&cp->c_data));
	else
	  fprintf (outstream, "\n");
	if (print_headers)
	  printf ("\n");
	break;
      }

  close_and_wait (outstream);
} /* simplestats_computer() */

computation *
simplestats_reader (int firstno, int lastno, char *lp)
{
  observable what;
  computation *clist, *cp;
  FILE *outstream;
  char *errstr;
  int j;
  arg_type args[] =
  {
    { NULL }
  };

  what = extract_observable (lp);
  if (what == wrong_o)
    return (NULL);
  lp = next_token (lp);

  errstr = lineparser_with_ostream (lp, args, &outstream);
  if (errstr != NULL)
    return_error (errstr);

  /* For each station allocate a computation structure and add it at
     the tail of the computation list. */
  clist = NULL;
  for (j = firstno; j <= lastno; j++)
    {
      add_collection (firstno, lastno, what);
      cp =  append_computation (&simplestats_computer, j, what, &clist);
      cp->args.simplestats.ostream = outstream;
    } /* for stations from firstno to lastno */

  return (clist);
} /* simplestats_reader () */

/*
 * Print history for observable data.
 *
 * Syntax is: listing <observable> [type=<type>] [every=<n>]
 *	being <aggregate> the string "aggregate", "compact" or
 *	"normal" (the default).  If "normal", the listing of the
 *	history values will be printed, one every <n>, which defaults
 *	to 1.  If "aggregate", the listing of the sum of the history
 *	values, grouped <n> by <n> is printed.  If "compact", the
 *	history values are printed in two columns: the first is the
 *	number of consecutive samples of the same value, the second is
 *	their value.
 */
static void
ds_listing (cm_args *argsp, unsigned long hlen, dshort *items)
{
  int j;
  int every = argsp->listing.every;
  FILE *outstream = argsp->listing.ostream;

  if (argsp->listing.compact)
    {
      int n = items[0].n;
      int val = items[0].val;

      for (j = 1; j < hlen; j++)
	{
	  if (val != items[j].val)
	    {
	      fprintf (outstream, "%d\t%d\n", n, val);
	      n = items[j].n;
	      val = items[j].val;
	    }
	  else
	    n += items[j].n;
	}
      fprintf (outstream, "%d\t%d\n", n, val);
    } /* compact */
  else if (argsp->listing.aggregate)
    {
      int rest = 0;
      int sum = 0;
      for (j = 0; j < hlen; j++)
	{
	  if (rest + items[j].n >= every)
	    {
	      fprintf (outstream, "%d\n", sum + (every-rest)*items[j].val);
	      for (rest += items[j].n - every; rest >= every; rest -= every)
		fprintf (outstream, "%d\n", every * items[j].val);
	      sum = rest * items[j].val;
	    }
	  else
	    {
	      rest += items[j].n;
	      sum += items[j].n * items[j].val;
	    }
	}
    } /* aggregate */
  else				/* one every group (no aggregate) */
    {
      int rest = 0;
      for (j = 0; j < hlen; j++)
	for (rest += items[j].n; rest >= every; rest -= every)
	  fprintf (outstream, "%d\n", items[j].val);
    }
  if (print_headers)
    printf ("\n");
} /* ds_listing() */

static void
f_listing (cm_args *argsp, unsigned long hlen, float *items)
{
  int j;
  int every = argsp->listing.every;
  FILE *outstream = argsp->listing.ostream;

  if (argsp->listing.compact)
    {
      int n = 1;
      float val = items[0];

      for (j = 1; j < hlen; j++)
	{
	  if (val != items[j])
	    {
	      fprintf (outstream, "%d\t%.4g\n", n, val);
	      n = 1;
	      val = items[j];
	    }
	  else
	    n += 1;
	}
      fprintf (outstream, "%d\t%.4g\n", n, val);
    } /* compact */

  else if (argsp->listing.aggregate)
    {
      double sum = 0;
      int k;

      j = 0;
      while (j < hlen)
	{
	  for (k = 0, sum = 0; k < every && j < hlen; k++, j++)
	    sum += items[j];
	  fprintf (outstream, "%.4g\n", sum);
	}
    } /* aggregate */

  else				/* one every group (no aggregate) */
    {
      for (j = every-1; j < hlen; j += every)
	fprintf (outstream, "%.4g\n", items[j]);
    }

  if (print_headers)
    printf ("\n");
} /* f_listing() */

static void
listing_computer (station *stp, observable what, cm_args *argsp)
{
  history *hp;
  FILE *outstream = argsp->listing.ostream;

  for (hp = stp->history_list; hp != NULL; hp = hp->h_next)
    if (hp->h_what == what)
      {
	unsigned long hlen = hp->h_current - hp->h_items;
	int every = argsp->listing.every;
	logical aggregate = argsp->listing.aggregate;
	logical compact = argsp->listing.compact;

        if (print_headers)
	  {
	    printf ("List of `%s' samples for station #%ld",
		   observable_string (what), stp->number);
	    if (compact)
	      {
	        printf (" (number and value of samples):\n");
	      }
	    else
	      {
	        if (every == 1)
	          printf (":\n");
	        else
	          {
		    if (aggregate)
		      printf (" (aggregates of %d):\n", every);
		    else
		      printf (" (every %d):\n", every);
	          }
	      }
	  }

	switch (what)
	  {
	  case wrong_o:
	    internal_error ();
	  case s_trudelay_o: case v_trudelay_o: case i_trudelay_o: case b_trudelay_o:
	    ds_listing (argsp, hlen, &hp->h_items->ds);
	    break;
	  default:
	    f_listing (argsp, hlen, &hp->h_items->f);
	    break;
	  }
      }

  close_and_wait (outstream);
} /* listing_computer () */

computation *
listing_reader (int firstno, int lastno, char *lp)
{
  observable what;
  computation *cp, *clist;
  int every = 0;
  logical aggregate = FALSE, compact = FALSE;
  FILE *outstream;
  char *str_type, *errstr;
  int j;
  arg_type args[] =
  {
    { "type", string_t, optional },
    { "every", uns_t, optional },
    { NULL }
  };

  what = extract_observable (lp);
  if (what == wrong_o)
    return (NULL);
  lp = next_token (lp);

  errstr = lineparser_with_ostream (lp, args, &outstream);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("type", str_type);
  get_arg ("every", every);

  if (str_type == NULL || streq (str_type, "normal"))
    {
      if (every == 0)
	every = 1;
    }
  else
    {
      if (streq (str_type, "aggregate"))
	{
	  if (every == 0)
	    every = 1;
	  aggregate = TRUE;
	}
      else if (streq (str_type, "compact"))
	{
	  if (every != 0)
	    return_error ("cannot specify every with type `compact'");
	  compact = TRUE;
	}
      else
	return_error ("type must be `normal', `aggregate' or `compact'");
    }

  /* For each station allocate a computation structure and add it at
     the tail of the computation list. */
  clist = NULL;
  for (j = firstno; j <= lastno; j++)
    {
      add_history (firstno, lastno, what);
      cp =  append_computation (&listing_computer, j, what, &clist);
      cp->args.listing.every = every;
      cp->args.listing.aggregate = aggregate;
      cp->args.listing.compact = compact;
      cp->args.listing.ostream = outstream;
    } /* for stations from firstno to lastno */

  return (clist);
} /* listing_reader () */

/*
 * Print the Kth quantile of the history elements.
 *
 * Syntax is: quantile <observable> q=<k>
 *	where <k> is a floating point number in the interval [0; 1].
 */
static history_item *sorted_item;
static unsigned long sorted_items;

static void
f_quantile (station *stp, history *hp, observable what, cm_args *argsp)
{
  float kth_q;
  unsigned long k;
  FILE *outstream = argsp->quantile.ostream;

  unsigned long hlen = hp->h_current - hp->h_items;
  int nbytes = hlen * sizeof(float);

  if (hlen == 0)
    {
      if (print_headers)
	printf ("%.4go quantile of `%s' for station #%ld %s\n",
		argsp->quantile.q, observable_string (what),
		stp->number, "not computed: no samples");
      else
	printf ("%.4g\t\n", argsp->quantile.q);

    }
  else
    {
      ASSERT (hlen < sorted_items);
      memcpy (sorted_item, &hp->h_items->f, nbytes);

      k = hlen * argsp->quantile.q;
      kth_q = nrselect (k+1, hlen, &sorted_item->f-1);

      if (print_headers)
	printf ("%.4go quantile of `%s' for station #%ld is %.4g\n",
		argsp->quantile.q, observable_string (what),
		stp->number, kth_q);
      else
	printf ("%.4g\t%.4g\n", argsp->quantile.q, kth_q);
    }
} /* f_quantile() */

static int 
compar (const void *A, const void *B)
{
  int aval = ((dshort *)A)->val;
  int bval = ((dshort *)B)->val;
  return (aval - bval);
}

static void
ds_quantile (station *stp, history *hp, observable what, cm_args *argsp)
{
  unsigned short kth_q;
  int j;
  double k, samples;
  FILE *outstream = argsp->quantile.ostream;

  unsigned long hlen = hp->h_current - hp->h_items;
  int nbytes = hlen * sizeof(dshort);

  if (hlen == 0)
    {
      if (print_headers)
	printf ("%.4go quantile of `%s' for station #%ld %s\n",
		argsp->quantile.q, observable_string (what),
		stp->number, "not computed: no samples");
      else
	printf ("%.4g\t\n", argsp->quantile.q);

    }
  else
    {
      ASSERT (hlen < sorted_items);
      memcpy (sorted_item, &hp->h_items->ds, nbytes);

      qsort (sorted_item, hlen, sizeof(dshort), compar);

      for (j = 0, samples = 0; j < hlen; j++)
	samples += sorted_item[j].ds.n;
      k = samples * argsp->quantile.q;
      for (j = 0; j < hlen && k > 0; j++)
	k -= sorted_item[j].ds.n;
      if (j > 0)
	j -= 1;
      kth_q = sorted_item[j].ds.val;
      if (print_headers)
	printf ("%.4go quantile of `%s' for station #%ld is %d\n",
		argsp->quantile.q, observable_string (what),
		stp->number, kth_q);
      else
	printf ("%.4g\t%d\n", argsp->quantile.q, kth_q);
    }
} /* ds_quantile() */

static void
quantile_computer (station *stp, observable what, cm_args *argsp)
{
  history *hp;
  FILE *outstream = argsp->quantile.ostream;

  for (hp = stp->history_list; hp != NULL; hp = hp->h_next)
    if (hp->h_what == what)
      {
	switch (what)
	  {
	  case wrong_o:
	    internal_error ();
	  case s_trudelay_o: case v_trudelay_o: case i_trudelay_o: case b_trudelay_o:
            ds_quantile (stp, hp, what, argsp);
	    break;
	  default:
            f_quantile (stp, hp, what, argsp);
	    break;
	  }
      }

  close_and_wait (outstream);
} /* quantile_computer () */

computation*
quantile_reader (int firstno, int lastno, char *lp)
{
  observable what;
  computation *cp, *clist;
  FILE *outstream;
  float q;
  int items, j;
  char *errstr;
  arg_type args[] =
  {
    { "q", float_t, required },
    { NULL }
  };

  what = extract_observable (lp);
  if (what == wrong_o)
    return (NULL);
  lp = next_token (lp);

  errstr = lineparser_with_ostream (lp, args, &outstream);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("q", q);

  if (q < 0 || q > 1)
    return_error ("q must lay between 0 and 1");

  /* For each station allocate a computation structure and add it at
     the tail of the computation list. */
  clist = NULL;
  for (j = firstno; j <= lastno; j++)
    {
      add_history (firstno, lastno, what);
      cp = append_computation (&quantile_computer, j, what, &clist);
      cp->args.quantile.q = q;
      cp->args.quantile.ostream = outstream;
    } /* for stations from firstno to lastno */

  if (what == s_trudelay_o || what == v_trudelay_o
      || what == i_trudelay_o || what == b_trudelay_o)
    items = trudhistlen;
  else
    items = histlen;
  if (firstno == 0)
    items *= stno;
  if (items > sorted_items)
    {
      int nbytes = items * sizeof (*sorted_item);

      sorted_items = items;
      if (sorted_item == NULL)
	sorted_item = malloc (nbytes);
      else
	sorted_item = realloc (sorted_item, nbytes);
      if (sorted_item == NULL)
	no_memory (nbytes);
      /* memset ((char *)sorted_item, 0, nbytes); */
    }

  return (clist);
} /* quantile_reader () */

static float *cs;
static int cslen;

static void
update_cs_size (int catn)
{
  if (catn == 0)
    catn = max_cslen;
  if (catn > cslen)
    {
      if (cslen != 0)
        free (cs);
      cslen = catn;
      cs = allocate (cslen * sizeof (*cs));
    }
} /* update_cs_size() */

static void
f_categorise (station *stp, history *hp, int hlen, observable what,
	      cm_args *argsp)
{
  collection *cp;
  float min, max, step;
  logical found;
  int j, catn;
  FILE *outstream = argsp->categorise.ostream;

  found = FALSE;
  min = max = 0;		/* just to make the compiler quiet */

  for (cp = stp->collection_list; cp != NULL; cp = cp->c_next)
    if (cp->c_what == what)
      {
        min = st_min (&cp->c_data);
        max = st_max (&cp->c_data);
        found = TRUE;
        break;
      }
  if (!found)
    {
      history_item *itemp = NULL;
      min = max = hp->h_items[0].f;
      for (itemp = hp->h_items + 1; itemp < hp->h_current; itemp++)
	{
	  if (itemp->f > max)
	    max = itemp->f;
	  else if (itemp->f < min)
	    min = itemp->f;
	}
    }
  catn = argsp->categorise.catn;
  if (catn == 0)
    {
      catn = floor (max - min + 1 + 0.5);
      if (catn > max_cslen)
	catn = max_cslen / 25;
      else
	{
	  min -= 0.5;
	  max += 0.5;
	}
    }
  categorise (&hp->h_items->f, hlen, min, max, catn, cs);

  step = (max - min) / catn;

  if (print_headers)
    printf ("%d categories of `%s' for station #%ld:\n",
	    catn, observable_string (what), stp->number);
  for (j = 0; j < catn; j++)
    fprintf (outstream, "%.4g\t%.4g\n", min + step * (j+0.5), cs[j]);
  if (print_headers)
    printf ("\n");

} /* f_categorise () */

static void
ds_categorise (station *stp, history *hp, int hlen, observable what,
	       cm_args *argsp)
{
  collection *cp;
  float min, max, step, st;
  logical found;
  int j, c, catn;
  FILE *outstream = argsp->categorise.ostream;

  found = FALSE;
  min = max = 0;	/* just to make the compiler quiet */

  for (cp = stp->collection_list; cp != NULL; cp = cp->c_next)
    if (cp->c_what == what)
      {
	min = st_min (&cp->c_data);
	max = st_max (&cp->c_data);
	found = TRUE;
	break;
      }
  if (!found)
    {
      history_item *itemp = NULL;
      min = max = hp->h_items[0].ds.val;
      for (itemp = hp->h_items + 1; itemp < hp->h_current; itemp++)
	{
	  if (itemp->ds.val > max)
	    max = itemp->ds.val;
	  else if (itemp->ds.val < min)
	    min = itemp->ds.val;
	}
    }
  catn = argsp->categorise.catn;
  if (catn == 0)
    {
      catn = floor (max - min + 1 + 0.5);
      if (catn > max_cslen)
	catn = max_cslen / 25;
      else
	{
	  min -= 0.5;
	  max += 0.5;
	}
    }

  step = (max - min) / catn;

  /* Let's always increment the first category when min==max. */
  st = (step > DBL_MIN) ? step : 1;
  memset ((char *)cs, 0, catn * sizeof (*cs));
  for (j = 0; j < hlen; j++)
    {
      c = (hp->h_items[j].ds.val - min) / st;
      if (c >= 0 && c < catn)
	cs[c] += hp->h_items[j].ds.n;
      else if (c < 0)
	cs[0] += hp->h_items[j].ds.n;
      else
	cs[catn-1] += hp->h_items[j].ds.n;
    }
  if (print_headers)
    printf ("%d categories of `%s' for station #%ld:\n",
	    catn, observable_string (what), stp->number);
  for (j = 0; j < catn; j++)
    fprintf (outstream, "%.4g\t%.4g\n", min + step * (j+0.5), cs[j]);
  if (print_headers)
    printf ("\n");

} /* ds_categorise() */


/*
 * Print the listing of the number of history samples listed by categories.
 *
 * Syntax is: categorise <observable> [catn=<n>]
 *	where the positive integer <n> is the number of categories to
 *	consider.  If <n> is 0, then a number of categories equal to
 *	(max - min + 1) is used.  This is useful if it is known that
 *	all the values are integers.
 */
static void
categorise_computer (station *stp, observable what, cm_args *argsp)
{
  history *hp;
  FILE *outstream = argsp->quantile.ostream;

  for (hp = stp->history_list; hp != NULL; hp = hp->h_next)
    if (hp->h_what == what)
      {
        int hlen = hp->h_current - hp->h_items;
	switch (what)
	  {
	  case wrong_o:
	    internal_error ();
	  case s_trudelay_o: case v_trudelay_o: case i_trudelay_o: case b_trudelay_o:
            ds_categorise (stp, hp, hlen, what, argsp);
	    break;
	  default:
            f_categorise (stp, hp, hlen, what, argsp);
	    break;
	  }
      }

  close_and_wait (outstream);
} /* categorise_computer () */

computation*
categorise_reader (int firstno, int lastno, char *lp)
{
  observable what;
  computation *cp, *clist;
  unsigned int catn;
  FILE *outstream;
  int j;
  char *errstr;
  arg_type args[] =
  {
    { "catn", uns_t, optional },
    { NULL }
  };

  what = extract_observable (lp);
  if (what == wrong_o)
    return (NULL);
  lp = next_token (lp);

  errstr = lineparser_with_ostream (lp, args, &outstream);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("catn", catn);

  update_cs_size (catn);

  /* For each station allocate a computation structure and add it at
     the tail of the computation list. */
  clist = NULL;
  for (j = firstno; j <= lastno; j++)
    {
      add_history (firstno, lastno, what);
      cp = append_computation (&categorise_computer, j, what, &clist);
      cp->args.categorise.catn = catn;
      cp->args.categorise.ostream = outstream;
    } /* for stations from firstno to lastno */

  return (clist);
} /* categorise_reader () */

#define ABS(a)	(((a) < 0) ? -(a) : (a))

static char **screen;
static int screen_rows, screen_cols;

static void
update_screen_size (int rows, int cols)
{
  int j;

  ASSERT (rows > 0 && cols > 0);
  if (rows > screen_rows || cols > screen_cols)
    {
      if (screen != NULL)
	{
	  for (j = 1; j < screen_rows+1; j++)
	    free (screen[j]);
	  free (screen);
	}
      if (rows > screen_rows) screen_rows = rows;
      if (cols > screen_cols) screen_cols = cols;
      screen = allocate ((screen_rows+1) * sizeof (char *));
      for (j = 1; j < screen_rows+1; j++)
	screen[j] = allocate ((screen_cols+2) * sizeof (char));
    }
} /* update_screen_size() */

static void
ds_ttydistrib (dshort items[], int hlen, float min, float max,
	       float avg, float sigma, int rows, int cols)
{
  float step, st;
  register int j, c;

  step = (max - min) / cols;

  /* Let's always increment the first category when min==max. */
  st = (step > DBL_MIN) ? step : 1;
  memset ((char *)cs, 0, cols * sizeof (*cs));
  for (j = 0; j < hlen; j++)
    {
      c = (items[j].val - min) / st;
      if (c >= 0 && c < cols)
        cs[c] += items[j].n;
      else if (c < 0)
        cs[0] += items[j].n;
      else
        cs[cols-1] += items[j].n;
    }
  ttygraph (cs, rows, cols, screen, TRUE);

  #if 0				/* one should untabify first */
  /* +/-3sigma marks */
  for (j = -3; j <= 3; j++)
    {
      col = (avg - min + j*sigma) / (max-min) * cols;
      if (col >= 0 && col < cols)
	screen[rows][col] = '0' + ABS(j);
    }
  #endif
} /* ds_ttydistrib () */

static void
f_ttydistrib (float items[], int hlen, float min, float max,
	      float avg, float sigma, int rows, int cols)
{
  float step, st;
  register int j, c;

  step = (max - min) / cols;

  /* Let's always increment the first category when min==max. */
  st = (step > DBL_MIN) ? step : 1;
  memset ((char *)cs, 0, cols * sizeof (*cs));
  for (j = 0; j < hlen; j++)
    {
      c = (items[j] - min) / st;
      if (c >= 0 && c < cols)
        cs[c] += 1;
      else if (c < 0)
        cs[0] += 1;
      else
        cs[cols-1] += 1;
    }
  ttygraph (cs, rows, cols, screen, TRUE);

  #if 0				/* one should untabify first */
  /* +/-3sigma marks */
  for (j = -3; j <= 3; j++)
    {
      col = (avg - min + j*sigma) / (max-min) * cols;
      if (col >= 0 && col < cols)
	screen[rows][col] = '0' + ABS(j);
    }
  #endif
} /* f_ttydistrib () */

/*
 * Print a graph of the distribution of the observable,
 * displayable on a tty.
 *
 * Syntax is: ttydistrib <observable> [rows=<rows>] [cols=<cols>]
 *	where <rows> and <cols> default to 22 and 78 respectively.
 */
static void
ttydistrib_computer (station *stp, observable what, cm_args *argsp)
{
  history *hp;
  collection *cp;
  int j, rows, cols;
  FILE *outstream = argsp->ttydistrib.ostream;

  for (hp = stp->history_list; hp != NULL; hp = hp->h_next)
    if (hp->h_what == what)
      for (cp = stp->collection_list; cp != NULL; cp = cp->c_next)
	if (cp->c_what == what)
	  {
	    int hlen = hp->h_current - hp->h_items;
	    double samples = st_n (&cp->c_data);
	    float min = st_min (&cp->c_data);
	    float max = st_max (&cp->c_data);
	    float avg, sigma;

	    if (samples < 2)
	      {
		fprintf (outstream,
			 "No plot with less than two `%s' samples.\n",
			 observable_string (what));
		return;
	      }

	    avg = st_avg (&cp->c_data);
	    sigma = st_stddev (&cp->c_data);
	    rows = argsp->ttydistrib.rows;
	    cols = argsp->ttydistrib.cols;

	    fprintf (outstream,
		     "Graph of `%s' distribution for station #%ld.\n",
		     observable_string (what), stp->number);
	    fprintf (outstream, "%10s  %13s  %13s  %13s  %13s\n",
		     "samples", "minimum", "maximum", "mean", "standard dev.");
	    fprintf (outstream, "%10.0f  %13g  %13g  %13g  %13g\n",
		     samples, min, max, avg, sigma);

	    switch (what)
	      {
	      case wrong_o:
		internal_error ();
	      case s_trudelay_o: case v_trudelay_o: case i_trudelay_o: case b_trudelay_o:
		ds_ttydistrib (&hp->h_items->ds, hlen, min, max, avg, sigma, rows, cols);
	        break;
	      default:
		f_ttydistrib (&hp->h_items->f, hlen, min, max, avg, sigma, rows, cols);
	    	break;
	      }

	    for (j = rows; j > 0; j--)
	      fprintf (outstream, screen[j]);
	    if (print_headers)
	      printf ("\n");
	    return;
	  }

  close_and_wait (outstream);
} /* ttydistrib_computer () */

computation*
ttydistrib_reader (int firstno, int lastno, char *lp)
{
  observable what;
  computation *cp, *clist;
  char *errstr;
  int rows, cols, j;
  FILE *outstream;
  arg_type args[] =
  {
    { "rows", uns_t, optional },
    { "cols", uns_t, optional },
    { NULL }
  };

  what = extract_observable (lp);
  if (what == wrong_o)
    return (NULL);
  lp = next_token (lp);

  errstr = lineparser_with_ostream (lp, args, &outstream);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("rows", rows);
  get_arg ("cols", cols);

  if (rows <= 2)
    rows = DEFAULT_ROWS;
  if (cols == 0)
    cols = DEFAULT_COLS;

  update_screen_size (rows, cols);
  update_cs_size (cols);

  /* For each station allocate a computation structure and add it at
     the tail of the computation list. */
  clist = NULL;
  for (j = firstno; j <= lastno; j++)
    {
      add_history (firstno, lastno, what);
      add_collection (firstno, lastno, what);
      cp = append_computation (&ttydistrib_computer, j, what, &clist);
      cp->args.ttydistrib.rows = rows;
      cp->args.ttydistrib.cols = cols;
      cp->args.ttydistrib.ostream = outstream;
    } /* for stations from firstno to lastno */

  return (clist);
} /* ttydistrib_reader () */

static void
ds_ttygraph (dshort items[], int hlen, double samples,
	     float min, float max, float avg, float sigma, int rows, int cols)
{
  int group, rest, save_rest, sum;
  register int j, k;

  rest = save_rest = sum = 0;
  k = 0;

  group = ceil (samples / cols);
  cols = floor (samples / group);

  for (j = 0; j < hlen && k < cols; j++)
    {
      if (rest + items[j].n >= group)
        {
      	  cs[k++] = sum + (group-rest) * items[j].val;
      	  for (rest += items[j].n - group; rest >= group; rest -=group )
              cs[k++] = group * items[j].val;
          sum = rest * items[j].val;
        }
      else
        {
      	  rest += items[j].n;
      	  sum += items[j].n * items[j].val;
    	}
    }

  ttygraph (cs, rows, cols, screen, FALSE);

} /* ds_ttygraph () */

static void
f_ttygraph (float items[], int hlen, float min, float max, float avg,
            float sigma, int rows, int cols)
{
  float sum;
  register int j, k, bin, group;

  group = excess_intdiv (hlen, cols);
  cols = hlen / group;
  hlen = cols * group;
  j = bin = sum = 0;

  while (j < hlen)
    {
      for (k = 0, sum = 0; k < group && j < hlen; k++, j++)
        sum += items[j];
      cs[bin++] = sum;
    }
  ttygraph (cs, rows, cols, screen, FALSE);

} /* f_ttygraph () */

/*
 * Print a graph of the observable versus the frame number,
 * displayable on a tty.
 *
 * Syntax is: ttygraph <observable> [rows=<rows>] [cols=<cols>]
 *	where <rows> and <cols> default to 22 and 78 respectively.
 */
static void
ttygraph_computer (station *stp, observable what, cm_args *argsp)
{
  history *hp;
  collection *cp;
  int j;
  FILE *outstream = argsp->ttygraph.ostream;

  for (hp = stp->history_list; hp != NULL; hp = hp->h_next)
    if (hp->h_what == what)
      for (cp = stp->collection_list; cp != NULL; cp = cp->c_next)
	if (cp->c_what == what)
	  {
	    int hlen = hp->h_current - hp->h_items;
	    double samples = st_n (&cp->c_data);
	    float min = st_min (&cp->c_data);
	    float max = st_max (&cp->c_data);
	    float avg, sigma;
	    int rows, cols;

	    if (samples < 2)
	      {
		fprintf (outstream,
			 "No plot with less than two `%s' samples.\n",
			 observable_string (what));
		return;
	      }

	    avg = st_avg (&cp->c_data);
	    sigma = st_stddev (&cp->c_data);
	    rows = argsp->ttygraph.rows;
	    cols = argsp->ttygraph.cols;

	    fprintf
	      (outstream,
	   "Graph of %.0f `%s' samples for station #%ld (min=%.4g, max=%.4g)\n",
	       samples, observable_string (what), stp->number, min, max);

	    switch (what)
	      {
	      case wrong_o:
		internal_error ();
	      case s_trudelay_o: case v_trudelay_o:
	      case i_trudelay_o: case b_trudelay_o:
		ds_ttygraph (&hp->h_items->ds, hlen, samples, min, max,
			     avg, sigma, rows, cols);
		break;
	      default:
		f_ttygraph
		  (&hp->h_items->f, hlen, min, max, avg, sigma, rows, cols);
		break;
	      }

	    for (j = rows; j > 0; j--)
	      fprintf (outstream, screen[j]);
	    if (print_headers)
	      printf ("\n");
	    return;
	  }

  close_and_wait (outstream);
} /* ttygraph_computer () */


computation*
ttygraph_reader (int firstno, int lastno, char *lp)
{
  observable what;
  computation *clist, *cp;
  char *errstr;
  int rows, cols, j;
  FILE *outstream;
  arg_type args[] =
  {
    { "rows", uns_t, optional },
    { "cols", uns_t, optional },
    { NULL }
  };

  what = extract_observable (lp);
  if (what == wrong_o)
    return (NULL);
  lp = next_token (lp);

  errstr = lineparser_with_ostream (lp, args, &outstream);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("rows", rows);
  get_arg ("cols", cols);

  if (rows <= 2)
    rows = DEFAULT_ROWS;
  if (cols <= 2)
    cols = DEFAULT_COLS;

  update_screen_size (rows, cols);
  update_cs_size (cols);

  /* For each station allocate a computation structure and add it at
     the tail of the computation list. */
  clist = NULL;
  for (j = firstno; j <= lastno; j++)
    {
      add_history (firstno, lastno, what);
      add_collection (firstno, lastno, what);
      cp = append_computation (&ttygraph_computer, j, what, &clist);
      cp->args.ttygraph.rows = rows;
      cp->args.ttygraph.cols = cols;
      cp->args.ttygraph.ostream = outstream;
    } /* for stations from firstno to lastno */

  return (clist);
} /* ttygraph_reader () */

void
free_computers_structures ()
{
  int j;

  if (cs != NULL)
    free (cs);
  cslen = 0;
  cs = NULL;

  if (screen != NULL)
    {
      for (j = 1; j < screen_rows+1; j++)
	free (screen[j]);
      free (screen);
    }
  screen_rows = screen_cols = 0;
  screen = NULL;

  if (sorted_item != NULL)
    free (sorted_item);
  sorted_items = 0;
  sorted_item = NULL;
} /* free_computers_structures() */
