/****************************************************************************

  $Source: /home/foda/fracas/RCS/fracas.c $
  $Revision: 7.15 $
  $Date: 1998/11/11 18:33:54 $
  $State: Exp $
  $Locker:  $

****************************************************************************/
#ident "$Id: fracas.c 7.15 1998/11/11 18:33:54 pot Exp $"

/*
 *  FRACAS stands for FRAmed Channel Access Simulator.
 */

#include "fracas.h"


logical checkonly;		/* don't run the emulation */
logical emulateonly;		/* don't make any computations */
logical kill_warnings;		/* don't print warnings */
logical printsrc;		/* print the source file */
logical terse;			/* don't print headers on stdout */
logical verbose;		/* show computation progress */

int stno;			/* number of stations */
int frametime;			/* duration of a frame */
int framesize;			/* capacity of the frame */
int rttime;			/* round trip time */
int histlen;			/* history length */
int trudhistlen;		/* TRU delay history length */
int max_cslen;			/* max length used for catn=0 */
int cyclelen;			/* length of the big table (frames) */
int warm_up_time;		/* initial emulation time without stats */
unsigned int global_seed;	/* for the generators who do not have a seed */
float reference_traffic;	/* global reference value for traffic */

char errmsg [BUFSIZ];		/* error messages buffer */
int lineno;			/* current input file line number */
station *st_table;		/* the stations array */
int curtime;			/* BOF of the current frame */
int curframeno;			/* frame number, starting from 1 */
int all_delay;			/* # of frames from curr_f to all_f */
long int totalloc;		/* total bytes allocated */
logical warmed;			/* curtime >= warm_up_time */
static random_state global_rs;	/* global random state */

initer *(*initer_reader)();	/* reader of the initer function */
char *initer_line;		/* initer argument line */
int initer_lineno;		/* initer argument line number */
initer *init_station;		/* pointer to initer function */
requester *(*requester_reader)(); /* reader of the initer function */
char *requester_line;		/* initer argument line */
int requester_lineno;		/* initer argument line number */
requester *compute_request;	/* pointer to requester function */
allocator *(*allocator_reader)(); /* reader of the initer function */
char *allocator_line;		/* initer argument line */
int allocator_lineno;		/* initer argument line number */
allocator *compute_allocation;	/* pointer to allocator function */
stopper *(*stopper_reader)();	/* reader of the initer function */
char *stopper_line;		/* initer argument line */
int stopper_lineno;		/* initer argument line number */
stopper *stop_condition;	/* pointer to stopper function */
computation *computation_list;	/* list of computations to do */

observable_name observable_names [] =
{
  { s_input_o, "s_input" },
  { v_input_o, "v_input" },
  { i_input_o, "i_input" },
  { b_input_o, "b_input" },
  { s_queue_o, "s_queue" },
  { v_queue_o, "v_queue" },
  { i_queue_o, "i_queue" },
  { b_queue_o, "b_queue" },
  { s_inbuffer_o, "s_inbuffer" },
  { v_inbuffer_o, "v_inbuffer" },
  { i_inbuffer_o, "i_inbuffer" },
  { b_inbuffer_o, "b_inbuffer" },
  { s_sent_o, "s_sent" },
  { v_sent_o, "v_sent" },
  { i_sent_o, "i_sent" },
  { b_sent_o, "b_sent" },
  { s_request_o, "s_request" },
  { v_request_o, "v_request" },
  { i_request_o, "i_request" },
  { b_request_o, "b_request" },
  { s_alloc_o, "s_allocation" },
  { v_alloc_o, "v_allocation" },
  { i_alloc_o, "i_allocation" },
  { b_alloc_o, "b_allocation" },
  { v_extra_o, "v_extraspace" },
  { i_extra_o, "i_extraspace" },
  { b_extra_o, "b_extraspace" },
  { b_unused_o, "b_unused" },
  { s_delay_o, "s_delay" },
  { v_delay_o, "v_delay" },
  { i_delay_o, "i_delay" },
  { b_delay_o, "b_delay" },
  { s_trudelay_o, "s_trudelay" },
  { v_trudelay_o, "v_trudelay" },
  { i_trudelay_o, "i_trudelay" },
  { b_trudelay_o, "b_trudelay" },
  { LAST }
};

#define COMMENT   '#'
#define PARAMETER '$'
#define CHARQUOTE '\\'
#define QUOTE     '"'

static char *progname;		/* pointer to program name */
static char *infilename;	/* pointer to input file name */
static logical endinput;	/* stop reading and emulate */
#define SUBSTPARS 10		/* number of possible substitutable params */
static char *substpar[SUBSTPARS]; /* substitutable parameters */
static int substlen[SUBSTPARS];	/* substitutable parameter lengths */


static void
input_error (char *message, int code)
{
  fprintf (stderr, "%s:%d: %s\n", infilename, lineno, message);
  exit (code);
}

void
input_warning (char *message)
{
  if (!kill_warnings)
    fprintf (stderr, "%s:%d: warning: %s\n", infilename, lineno, message);
}

void
runtime_error (char *message)
{
  fprintf (stderr, "%s: in %s: %s\n", progname, infilename, message);
  fflush (stderr);
}

void
_internal_error (char *file, int line)
{
  fprintf (stderr, "%s: internal error [%s:%d] -- ",
	   progname, file, line);
  abort ();
}

#ifndef DEBUG
void *
allocate (int n)
{
  void *ptr;

  if (n < 0)
    internal_error();
  if (n == 0)
    return (NULL);
  ptr = calloc (n, 1);
  if (ptr == NULL)
    no_memory (n);
  totalloc += n;
  return (ptr);
}
#endif

void
no_memory (int n)
{
  char error[128];
  sprintf (error,
	   "can't allocate %d KB (%ld KB already allocated)",
	   n/1024+1, totalloc/1024+1);
  input_error (error, 4);
}

static void
usage (void)
{
  fprintf (stderr, "usage: %s [options] [-N <parameter>] files\n", progname);
  fprintf (stderr, "Where options are:\n\
	-N	N is a digit, <parameter> is a string.\n\
	-c	checkonly - don't run the emulation after the init phase\n\
	-e	print all known values on stdout, skip all computations\n\
	-s	print the entire source file before the computation results\n\
	-t	terse - do not print headers on standard output\n\
	-v	verbose - print messages showing the process of computations\n\
	-w	warnings - don't print warnings\n\
");
  return;
}

/*
 * Close STREAM and wait for all possible children.
 *
 * Use this only for closing ouput pipes, and only one at a time: no
 * concurrent childs!
 * This must be done because pclose(3) is broken.  Indeed, it first
 * closes the output stream, without checking if it was opened by
 * popen.  Next it waits.  If the stream was not opened by popen, the
 * process is blocked until there are no running childs (possibly
 * forever).  Even worse, if childs A and B are created with popen, then
 * one pclose's A and B finishes first, when you then pclose B, pclose
 * never returns!
 */
void
close_and_wait (FILE *stream)
{
  if (stream == stdout)
    fflush (stdout);
  else
    pclose (stream);
}

/*
 * Return the name of the observable.
 */
char *
observable_string (observable what)
{
  observable_name *on;

  for (on = observable_names; on->what != LAST; on++)
    if (on->what == what)
      return (on->name);
  internal_error ();
  return (NULL);		/* just to make the compiler quiet */
}

static int
observable_value (frame *fp, observable what)
{
  switch (what)
    {
    case s_input_o: return (fp->input.s);
    case v_input_o: return (fp->input.v);
    case i_input_o: return (fp->input.i);
    case b_input_o: return (fp->input.b);

    case s_queue_o: return (fp->queue.s);
    case v_queue_o: return (fp->queue.v);
    case i_queue_o: return (fp->queue.i);
    case b_queue_o: return (fp->queue.b);

    case s_inbuffer_o: return (fp->queue.s + fp->sent.s);
    case v_inbuffer_o: return (fp->queue.v + fp->sent.v);
    case i_inbuffer_o: return (fp->queue.i + fp->sent.i);
    case b_inbuffer_o: return (fp->queue.b + fp->sent.b);

    case s_sent_o: return (fp->sent.s);
    case v_sent_o: return (fp->sent.v);
    case i_sent_o: return (fp->sent.i);
    case b_sent_o: return (fp->sent.b);

    case s_request_o: return (fp->request.s);
    case v_request_o: return (fp->request.v);
    case i_request_o: return (fp->request.i);
    case b_request_o: return (fp->request.b);

    case s_alloc_o: return (fp->allocation.s);
    case v_alloc_o: return (fp->allocation.v);
    case i_alloc_o: return (fp->allocation.i);
    case b_alloc_o: return (fp->allocation.b);

    case v_extra_o: return (fp->extraspace.v);
    case i_extra_o: return (fp->extraspace.i);
    case b_extra_o: return (fp->extraspace.b);

    case b_unused_o: return (fp->allocation.b + fp->extraspace.b - fp->sent.b);

    default:
      internal_error ();
      return (0);		/* to let the compiler quiet */
    }
} /* observable_value() */



/*
 * Return a pointer to a space of size strlen(cp)+1 allocated
 * with allocate() where the string CP has been copied.
 */
char *
savestr (char *cp)
{
  if (cp == NULL)
    return (NULL);
  return savenstr (cp, strlen (cp));
}

/*
 * Return a pointer to a space of size LEN+1 allocated with allocate() where
 * the string CP has been copied for at most the first LEN characters.
 */
char *
savenstr (char *cp, int len)
{
  char *dp = allocate (len + 1);
  strncpy (dp, cp, len);
  dp[len] = '\0';
  return dp;
}

/* Return a newly-allocated string whose contents
   concatenate those of s1, s2, s3.  */
char *
concat3 (char *s1, char *s2, char *s3)
{
  int len1 = strlen (s1);
  int len2 = strlen (s2);
  int len3 = strlen (s3);
  char *result = allocate (len1 + len2 + len3 + 1);

  strcpy (result, s1);
  strcpy (result + len1, s2);
  strcpy (result + len1 + len2, s3);
  result[len1 + len2 + len3] = '\0';

  return result;
}


/*
 * Read a line from stdin in BUF discarding leading spaces and tabs,
 * and discarding characters starting from the MAXth up to a newline
 * or a COMMENT character.  Delete the ending newline or COMMENT
 * character and terminate the line with a null char.  Then substitute
 * the $N strings (with N being a digit) with the relevant
 * substitutable parameter.
 * BUF is assumed to be MAX characters long at least.  Return the
 * length of the string put in BUF, not including the terminating null
 * character.  If the string put in BUF has null lenght, and there was
 * not any comment in the line read, then return -1.
 */
static int
readline (FILE * infile, char *buf, int max)
{
  int j, c;
  logical charquote, substitute;

  lineno += 1;
  j = 0;
  charquote = substitute = FALSE;
  do
    c = getc (infile);
  while (c == ' ' || c == '\t' || c == '\f');
  do
    {
      if (c == CHARQUOTE && !charquote)
	{
	  charquote = TRUE;
	  continue;
	}
      else if (c == PARAMETER && !charquote)
	{
	  substitute = TRUE;
	  continue;
	}
      else if (c == EOF || c == '\n')
	{
	  if (charquote)
	    {
	      charquote = FALSE;
	      lineno += 1;
	      continue;
	    }
	  buf[j] = '\0';
	  if (j == 0)
	    j = -1;
	  break;
	}
      else if (c == COMMENT && !charquote)
	{
	  buf[j] = '\0';
	  do
	    c = getc (infile);
	  while (c != EOF && c != '\n');
	  break;
	}
      else if (charquote)
	{
	  buf[j++] = CHARQUOTE;
	}
      charquote = FALSE;

      if (substitute && c >= '0' && c <= '9')
	{
	  int parm = c - '0';

	  if (substlen[parm] < 0)
	    input_error ("substitutable parameter not initialised", 1);
	  if (j+substlen[parm] >= max)
	    input_error
	      ("input line with substituted parameters is too long", 3);
	  strcpy (&buf[j], substpar[parm]);
	  j += substlen[parm];
	  substitute = FALSE;
	}
      else
	{
	  if (j >= max)
	    input_error ("input line is too long", 3);
	  buf[j++] = c;
	}
    } while (c = getc (infile), TRUE);
  return (j);
} /* readline() */



/*
 * first_token_eq returns TRUE if LINE's first token is equal to TOK.
 */
logical
first_token_eq (char *line, char *tok)
{
  if (tok == NULL || line == NULL)
    return (FALSE);
  while (*tok)
    if (*line++ != *tok++)
      return (FALSE);
  if (*line && strchr (TOKS, *line))
    return (FALSE);
  return (TRUE);
} /* first_token_eq() */

char *
next_token (char *lp)
{
  if (lp == NULL)
    return (NULL);
  if (lp[0] == '"') do
    {
      lp = strchr(lp+1, '"');
      if (lp == NULL)
	return (NULL);
    }
  while (lp[-1] == CHARQUOTE);
  lp = strpbrk (lp, SEPS);
  if (lp == NULL)
    return (NULL);
  return (strpbrk (lp, TOKS));
} /* next_token() */

static char *
first_string (char *lp)
{
  if (lp == NULL)
    return (savestr (""));
  if (lp[0] == '"')
    {
      char *str = allocate (strlen (lp));
      char *f = lp+1;
      char *t = str;

      while (*f != '\0')
	if (*f == '"')
	  {
	    *t = '\0';
	    return (str);
	  }
	else
	  {
	    if (*f == CHARQUOTE)
	      f += 1;
	    *t++ = *f++;
	  }
      return (NULL);		/* error: unterminated quoted string */
    }
  else if (strpbrk (lp, SEPS) != NULL)
    return (savenstr (lp, strpbrk (lp, SEPS) - lp));
  else
    return (savestr (lp));
} /* first_string() */

static FILE *
smart_fopen (char *filename, char *type)
{
  FILE *result;
  char *cp;
  int c;

  cp = strrchr (filename, '.');

  if (cp == NULL
      || (!streq (cp, ".gz") && !streq (cp, ".Z") && !streq (cp, ".z")))
    result = fopen (filename, type);
  else if (streq (type, "r"))
    {
      if (streq (cp, ".gz"))
	result = popen (concat3 ("gzip -cd ", filename, ""), "r");
      else if (streq (cp, ".Z"))
	result = popen (concat3 ("compress -cd ", filename, ""), "r");
      else
	result = popen (concat3 ("zcat ", filename, ""), "r");
    }
  else
    {
      /* To be implemented. */
      internal_error ();
      return (NULL);		/* to make the compiler quiet */
    }

  if (result == NULL)
    {
      sprintf (errmsg, "%s -- %s", filename, sys_errlist[errno]);
      return (NULL);
    }
  c = getc (result);
  if (c == EOF)
    {
      sprintf (errmsg, "%s: decompression error", filename);
      return (NULL);
    }
  ungetc (c, result);
  return (result);
} /* smart_fopen() */

/*
 * This function works as a universal reader.  It takes arguments LP that
 * points to the current input line and ARGS, that points to a NULL
 * terminated array of arg_type's.  The input line is parsed according
 * to ARGS and the elements of args are filled with the proper values.
 * An error message is returned if any errors are encountered, or NULL
 * if no errors.
 */
char *
lineparser (char *lp, arg_type *args)
{
  arg_type *argsp;
  char *filename;
  logical found;

  if (lp != NULL)
    lp = strpbrk (lp, TOKS);
  while (lp != NULL)
    {
      found = FALSE;
      for (argsp = args; argsp->name != NULL; argsp++)
	if (tokenis (argsp->name))
	  {
	    lp += strlen (argsp->name);
	    if (*lp == '=')
	      lp += 1;
	    else
	      return (concat3 ("`=' expected after ", argsp->name, ""));
	    switch (argsp->type)
	      {
	      case int_t:
		if (sscanf (lp, "%i", &argsp->value.int_v) != 1)
		  return (concat3 ("argument `", argsp->name, "' missing"));
		break;

	      case uns_t:
		if (sscanf (lp, "%i", &argsp->value.uns_v) != 1)
		  return (concat3 ("argument `", argsp->name, "' missing"));
		break;

	      case float_t:
		if (sscanf (lp, "%g", &argsp->value.float_v) != 1)
		  return (concat3 ("argument `", argsp->name, "' missing"));
		break;

	      case log_t:
		if (tokenis ("y") || tokenis ("yes"))
		  argsp->value.log_v = TRUE;
		else if (tokenis ("n") || tokenis ("no"))
		  argsp->value.log_v = FALSE;
		else
		  return (concat3
			  ("argument `", argsp->name, "': y/n expected"));
		break;

	      case string_t:
		argsp->value.string_v = first_string (lp);
		if (argsp->value.string_v == NULL)
		  return ("closing `\"' expected");
		if (strlen (argsp->value.string_v) == 0)
		  return (concat3 ("argument `", argsp->name, "' missing"));
		break;

	      case ifile_t:
		filename = first_string (lp);
		if (filename == NULL)
		  return ("closing `\"' expected");
		if (strlen (filename) == 0)
		  return (concat3 ("argument `", argsp->name, "' missing"));
		argsp->value.stream_v = smart_fopen (filename, "r");
		if (argsp->value.stream_v == NULL)
		  return ((char *)sys_errlist[errno]);
		break;

	      case ofile_t:
		filename = first_string (lp);
		if (filename == NULL)
		  return ("closing `\"' expected");
		if (strlen (filename) == 0)
		  return (concat3 ("argument `", argsp->name, "' missing"));
		argsp->value.stream_v = fopen (filename, "w");
		if (argsp->value.stream_v == NULL)
		  return ((char *)sys_errlist[errno]);
		break;

	      case afile_t:
		filename = first_string (lp);
		if (filename == NULL)
		  return ("closing `\"' expected");
		if (strlen (filename) == 0)
		  return (concat3 ("argument `", argsp->name, "' missing"));
		argsp->value.stream_v = fopen (filename, "a");
		if (argsp->value.stream_v == NULL)
		  return ((char *)sys_errlist[errno]);
		break;

	      case ipipe_t:
		filename = first_string (lp);
		if (filename == NULL)
		  return ("closing `\"' expected");
		if (strlen (filename) == 0)
		  return (concat3 ("argument `", argsp->name, "' missing"));
		argsp->value.stream_v = popen (filename, "r");
		if (argsp->value.stream_v == NULL)
		  return ((char *)sys_errlist[errno]);
		break;

	      case opipe_t:
		filename = first_string (lp);
		if (filename == NULL)
		  return ("closing `\"' expected");
		if (strlen (filename) == 0)
		  return (concat3 ("argument `", argsp->name, "' missing"));
		argsp->value.stream_v = popen (filename, "w");
		if (argsp->value.stream_v == NULL)
		  return ((char *)sys_errlist[errno]);
		break;

	      case iopipe_t:
		/* Not yet implemented. */
		internal_error ();
	      } /* switch of <argsp->type> */

	    argsp->state = got;
	    found = TRUE;
	    break;
	  } /* for all the possible args if tokenis(argsp->name) */

      if (!found)
	{
	  char *equalp = strchr (lp, '=');
	  if (equalp != NULL)
	    *equalp = '\0';
	  return (concat3 ("parameter `", lp, "' unknown"));
	}

    lp = next_token (lp);
  } /* while there is something more to read */

  for (argsp = args; argsp->name != NULL; argsp++)
    if (argsp->state == required)
      return (concat3 ("parameter `", argsp->name, "' is missing"));

  return (NULL);
} /* lineparser() */

/*
 * This function works uses the universal reader above.  It takes the
 * same arguments plus OSTREAMP, the address of the variable where the
 * output stream is written.  The arguments expected on the input line
 * are those specified in the ARGS plus either "ofile", "afile" or
 * "opipe".  If more than one of these is specified or an error occurs
 * while reading one, an error string is returned.  If none is
 * specified, stdio is written in *OSTREAMP.  If lineparser returns an
 * error, that error is returned.
 */
char *
lineparser_with_ostream (char *lp, arg_type *inargs, FILE **ostream)
{
  char *errstr;
  FILE *ofile, *afile, *opipe;
  int inargs_len, osargs_len;
  arg_type *args;
  arg_type osargs[] =
  {
    { "ofile", ofile_t, optional },
    { "afile", afile_t, optional },
    { "opipe", opipe_t, optional },
    { NULL }
  };

  /* Count the number of entries in inargs and osargs.
     Allocate a new arg_type array args.
     Concatenate inargs and osargs into args. */
  for (inargs_len = 0; inargs[inargs_len].name != NULL; inargs_len++)
    continue;
  osargs_len = sizeof(osargs) / sizeof(arg_type);
  args = allocate ((inargs_len+osargs_len) * sizeof(arg_type));
  memcpy (args, inargs, inargs_len * sizeof(arg_type));
  memcpy (args+inargs_len, osargs, osargs_len * sizeof(arg_type));

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return (errstr);
  get_arg ("ofile", ofile);
  get_arg ("afile", afile);
  get_arg ("opipe", opipe);

  memcpy (inargs, args, inargs_len * sizeof(arg_type));
  free (args);

  if (ofile == NULL && afile == NULL && opipe == NULL)
    *ostream = stdout;
  else if (ofile != NULL)
    {
      if (afile == NULL && opipe == NULL)
	*ostream = ofile;
      else
	return ("only one of `opipe', `afile' or `ofile' must be specified");
    }
  else if (afile != NULL)
    {
      if (opipe == NULL)	/* here certainly ofile==NULL */
	*ostream = afile;
      else
	return ("only one of `opipe', `afile' or `ofile' must be specified");
    }
  else
    *ostream = opipe;

  return (NULL);
} /* lineparser_with_ostream() */

/*
 * This function put in *VARP the value of the arg_type structure with name
 * equal to NAME in a NULL terminated array ARGS of type_args structures.
 */
void
_get_value (arg_type *args, char *name, void *varp)
{
  arg_type *argp;

  for (argp = args; argp->name != NULL; argp++)
    if (streq (argp->name, name))
      break;
  if (argp->name == NULL)
    internal_error ();		/* no arg_type with given name found */
  switch (argp->type)
    {
    case int_t:
      *(int *)varp = argp->value.int_v; break;
    case uns_t:
      *(unsigned int *)varp = argp->value.uns_v; break;
    case float_t:
      *(float *)varp = argp->value.float_v; break;
    case string_t:
      *(char **)varp = argp->value.string_v; break;
    case log_t:
      *(logical *)varp = argp->value.log_v; break;
    case ifile_t: case ipipe_t:
    case ofile_t: case opipe_t:
    case afile_t:
    case iopipe_t:
      *(FILE **)varp = argp->value.stream_v; break;
    }
} /* _get_value() */

/*
 * read_global_vars: read the input file and parse the global section.
 */
static char *
read_global_vars (FILE * infile)
{
  char line[BUFSIZ], *lp;
  int len;


  /* Skip blank and comment lines. */
  do
    len = readline (infile, line, BUFSIZ);
  while (len < 1 && !feof (infile));

  /* Read consecutive nonblank lines. */
  do
    {
      if (len > 0)
	{
	  lp = line;
	  if (tokenis ("frametime"))
	    {
	      if (frametime != 0)
		return ("frametime already initalised");
	      if (sscanf (line, "%*s %u", &frametime) != 1)
		return ("cannot read frametime");
	      if (framesize <= 0)
		return ("framesize must be positive");
	    }
	  else if (tokenis ("framesize"))
	    {
	      if (framesize != 0)
		return ("framesize already initialised");
	      if (sscanf (line, "%*s %u", &framesize) != 1)
		return ("cannot read framesize");
	      if (framesize <= 0)
		return ("framesize must be positive");
	    }
	  else if (tokenis ("rttime"))
	    {
	      if (rttime != 0)
		return ("rttime already initialised");
	      if (sscanf (line, "%*s %u", &rttime) != 1)
		return ("cannot read rttime");
	      if (rttime <= 0)
		return ("rttime must be positive");
	    }
	  else if (tokenis ("histlen"))
	    {
	      if (histlen != 0)
		return ("histlen already initialised");
	      if (sscanf (line, "%*s %u", &histlen) != 1)
		return ("cannot read histlen");
	      if (histlen <= 0)
		return ("histlen must be positive");
	    }
	  else if (tokenis ("trudhistlen"))
	    {
	      if (trudhistlen != 0)
		return ("trudhistlen already initialised");
	      if (sscanf (line, "%*s %u", &trudhistlen) != 1)
		return ("cannot read trudhistlen");
	      if (trudhistlen <= 0)
		return ("trudhistlen must be positive");
	    }
	  else if (tokenis ("cyclelen"))
	    {
	      if (cyclelen != 0)
		return ("cyclelen already initialised");
	      if (sscanf (line, "%*s %u", &cyclelen) != 1)
		return ("cannot read cyclelen");
	      if (cyclelen <= 0)
		return ("cyclelen must be positive");
	    }
	  else if (tokenis ("warmup"))
	    {
	      if (warm_up_time != 0)
		return ("warmup already initialised");
	      if (sscanf (line, "%*s %u", &warm_up_time) != 1)
		return ("cannot read warmup");
	      if (warm_up_time < 0)
		return ("warm_up_time must be positive");
	    }
	  else if (tokenis ("seed"))
	    {
	      if (global_seed != 0)
		return ("seed already initialised");
	      if (sscanf (line, "%*s %u", &global_seed) != 1)
		return ("cannot read seed");
	    }
	  else if (tokenis ("max_cslen"))
	    {
	      if (max_cslen != 0)
		return ("max_cslen already initialised");
	      if (sscanf (line, "%*s %u", &max_cslen) != 1)
		return ("cannot read max_cslen");
	      if (max_cslen <= 0)
		return ("max_cslen must be positive");
	    }
	  else if (tokenis ("ref_traffic"))
	    {
	      if (reference_traffic != 0)
		return ("ref_traffic already initialised");
	      if (sscanf (line, "%*s %f", &reference_traffic) != 1)
		return ("cannot read ref_traffic");
	      if (reference_traffic <= 0)
		return ("ref_traffic must be positive");
	    }
	  else
	    {
	      return ("unknown global variable");
	    }
	} /* if (len > 0) */
      len = readline (infile, line, BUFSIZ);
    }
  while (len >= 0);

  /* Check that all nonoptional global variables have been read. */
  if (frametime == 0)
    return ("frametime not initialised");
  else if (framesize == 0)
    return ("framesize not initialised");
  else if (rttime == 0)
    return ("rttime not initialised");
  else if (histlen == 0)
    return ("histlen not initialised");

  /* More checks. */

  /* Initialise optional global variables. */
  if (trudhistlen == 0)
    trudhistlen = 2 * histlen;
  if (max_cslen == 0)
    max_cslen = 25000;
  if (cyclelen == 0)
    cyclelen = 4*rttime/frametime + 8;
  if (global_seed == 0)
    {
      char warn[BUFSIZ];
      global_seed = time (NULL);
      sprintf (warn, "global seed set to %u", global_seed);
      input_warning (warn);
    }

  /* Other initialisations. */
  initstate (global_seed, &global_rs, RND_STATE_SIZE);

  return (NULL);
} /* read_global_vars() */



/*
 * read_initers: read the input file and parse the global section.
 */
static char *
read_initers (FILE * infile)
{
  extern initer_n initers [];
  char line[BUFSIZ], *lp;
  int len;
  initer_n *initer_np;
  logical found;

  /* Skip blank and comment lines. */
  do
    len = readline (infile, line, BUFSIZ);
  while (len < 1 && !feof (infile));

  /* Read consecutive nonblank lines. */
  do
    {
      if (len > 0)
	{
	  lp = line;
	  if (!tokenis ("initer"))
	    return ("initer expected");
	  if (init_station != NULL)
	    return ("more than one initer specified");
	  lp = next_token (lp);
	  found = FALSE;
	  for (initer_np = initers; initer_np->name != NULL; initer_np++)
	    if (tokenis (initer_np->name))
	      {
		found = TRUE;
		initer_reader = initer_np->reader;
		initer_line = savestr (next_token (lp));
		initer_lineno = lineno;
		break;
	      }
	  if (!found)
	    return ("unknown initer");
	} /* if (len > 0) */
      len = readline (infile, line, BUFSIZ);
    }
  while (len >= 0);

  return (NULL);
} /* read_initers() */



/*
 * read_requesters: read the input file and parse the global section.
 */
static char *
read_requesters (FILE * infile)
{
  extern requester_n requesters [];
  char line[BUFSIZ], *lp;
  int len;
  requester_n *requester_np;
  logical found;


  /* Skip blank and comment lines. */
  do
    len = readline (infile, line, BUFSIZ);
  while (len < 1 && !feof (infile));

  /* Read consecutive nonblank lines. */
  do
    {
      if (len > 0)
	{
	  lp = line;
	  if (!tokenis ("requester"))
	    return ("requester expected");
	  if (compute_request != NULL)
	    return ("more than one requester specified");
	  lp = next_token (lp);
	  found = FALSE;
	  for (requester_np = requesters;
	       requester_np->name != NULL; requester_np++)
	    if (tokenis (requester_np->name))
	      {
		found = TRUE;
		requester_reader = requester_np->reader;
		requester_line = savestr (next_token (lp));
		requester_lineno = lineno;
		break;
	      }
	  if (!found)
	    return ("unknown requester type");
	} /* if (len > 0) */
      len = readline (infile, line, BUFSIZ);
    }
  while (len >= 0);

  return (NULL);
} /* read_requesters() */



/*
 * read_allocators: read the input file and parse the global section.
 */
static char *
read_allocators (FILE * infile)
{
  extern allocator_n allocators [];
  char line[BUFSIZ], *lp;
  int len;
  allocator_n *allocator_np;
  logical found;


  /* Skip blank and comment lines. */
  do
    len = readline (infile, line, BUFSIZ);
  while (len < 1 && !feof (infile));

  /* Read consecutive nonblank lines. */
  do
    {
      if (len > 0)
	{
	  lp = line;
	  if (!tokenis ("allocator"))
	    return ("allocator expected");
	  if (compute_allocation != NULL)
	    return ("more than one allocator specified");
	  lp = next_token (lp);
	  found = FALSE;
	  for (allocator_np = allocators;
	       allocator_np->name != NULL; allocator_np++)
	    if (tokenis (allocator_np->name))
	      {
		found = TRUE;
		allocator_reader = allocator_np->reader;
		allocator_line = savestr (next_token (lp));
		allocator_lineno = lineno;
		break;
	      }
	  if (!found)
	    return ("unknown allocator");
	} /* if (len > 0) */
      len = readline (infile, line, BUFSIZ);
    }
  while (len >= 0);

  return (NULL);
} /* read_allocators() */



/*
 * read_stoppers: read the input file and parse the global section.
 */
static char *
read_stoppers (FILE * infile)
{
  extern stopper_n stoppers [];
  char line[BUFSIZ], *lp;
  int len;
  stopper_n *stopper_np;
  logical found;


  /* Skip blank and comment lines. */
  do
    len = readline (infile, line, BUFSIZ);
  while (len < 1 && !feof (infile));

  /* Read consecutive nonblank lines. */
  do
    {
      if (len > 0)
	{
	  lp = line;
	  if (!tokenis ("stopper"))
	    return ("stopper expected");
	  if (stop_condition != NULL)
	    return ("more than one stopper specified");
	  lp = next_token (lp);
	  found = FALSE;
	  for (stopper_np = stoppers;
	       stopper_np->name != NULL; stopper_np++)
	    if (tokenis (stopper_np->name))
	      {
		found = TRUE;
		stopper_reader = stopper_np->reader;
		stopper_line = savestr (next_token (lp));
		stopper_lineno = lineno;
		break;
	      }
	  if (!found)
	    return ("unknown stopper");
	} /* if (len > 0) */
      len = readline (infile, line, BUFSIZ);
    }
  while (len >= 0);

  return (NULL);
} /* read_stoppers() */

/*
 * read_generator: parse the traffic generator data.
 */
static char *
read_generator (char *lp, int firstno, int lastno)
{
  extern sampler_n samplers [];
  sampler_n *sampler_np;
  sampler_r *sampler_rp;
  tt traffic;
  int j;

  lp = next_token (lp);
  if (tokenis ("s")) traffic = s;
  else if (tokenis ("v")) traffic = v;
  else if (tokenis ("i")) traffic = i;
  else if (tokenis ("b")) traffic = b;
  else
    return ("unknown traffic type");

  /* Find the correct sampler reader. */
  lp = next_token (lp);
  sampler_rp = NULL;
  for (sampler_np = samplers; sampler_np->name != NULL; sampler_np++)
    if (tokenis (sampler_np->name))
      {
	sampler_rp = sampler_np->reader;
	break;
      }
  lp = next_token (lp);
  if (sampler_rp == NULL)
    return ("unknown generator type");

  /* For all the stations we have just allocated, call the
     appropriate reader and add the returned generator at
     the head of the tg list. */
  for (j = firstno; j <= lastno; j++)
    {
      tg *tgp = sampler_rp (lp);
      if (tgp == NULL)
	return (errmsg);
      tgp->ttype = traffic;
      tgp->t_next = st_table[j].tg_list;
      st_table[j].tg_list = tgp;
    }

  return (NULL);
} /* read_generator() */



/*
 * Read the stream request of a station
 */
static char *
read_stream_request (char *lp, int firstno, int lastno)
{
  char *errstr;
  int j;
  arg_type args[] =
  {
    { "sreq", uns_t, optional },
    { NULL }
  };

  lp = next_token (lp);

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return (errstr);

  for (j = firstno; j <= lastno; j++)
    get_arg ("sreq", st_table[j].sreq);

  return (NULL);
} /* read_stream_request () */



/*
 * Read the vbr request of a station
 */
static char *
read_vbr_request (char *lp, int firstno, int lastno)
{
  char *errstr;
  int vminreq, vmaxreq, j;
  float Ji;
  arg_type args[] =
  {
    { "vminreq", uns_t, optional },
    { "vmaxreq", uns_t, optional },
    { "Ji", float_t, optional },
    { NULL }
  };

  /* Read the minimum and maximum vbr requests and the input jitter. */
  lp = next_token (lp);

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return (errstr);
  get_arg ("vminreq", vminreq);
  get_arg ("vmaxreq", vmaxreq);
  get_arg ("Ji", Ji);

  if (vmaxreq == 0)
    vmaxreq = vminreq;
  if (vmaxreq > 0 && vmaxreq < vminreq)
    return ("`vmaxreq' must be greater than `vminreq'");
  if (Ji < 0)
    return_error ("Ji is negative");

  for (j = firstno; j <= lastno; j++)
    {
      st_table[j].vminreq = vminreq;
      st_table[j].vmaxreq = vmaxreq;
      st_table[j].Ji = Ji / frametime;
    }

  return (NULL);
} /* read_vbr_request () */



/*
 * Read the maximum input queue length of a station
 */
static char *
read_maxqueuelen (char *lp, int firstno, int lastno)
{
  char *errstr;
  int j;
  arg_type args[] =
  {
    { "s", uns_t, optional },
    { "v", uns_t, optional },
    { "i", uns_t, optional },
    { "b", uns_t, optional },
    { NULL }
  };

  /* Read maximum input queue length */
  lp = next_token (lp);

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return (errstr);

  for (j = firstno; j <= lastno; j++)
    {
      get_arg ("s", st_table[j].maxqueuelen.s);
      get_arg ("v", st_table[j].maxqueuelen.v);
      get_arg ("i", st_table[j].maxqueuelen.i);
      get_arg ("b", st_table[j].maxqueuelen.b);
    }

  return (NULL);
} /* read_maxqueuelen() */



/*
 * Read the input file and parse the station data
 */
static char *
read_station (FILE * infile, char *lp)
{
  char *errstr;
  char line[BUFSIZ];
  int len, noread;
  int firstno, lastno;
  int j, nbytes;
  station *stp;


  /* Read nonblank lines */
  lastno = 0;
  if (lp == NULL
      || (noread = sscanf (lp, "%*s %d:%d", &firstno, &lastno)) < 1)
    return ("missing station number");
  if (noread == 1)
    lastno = firstno;

  /* Check firstno and lastno. */
  if (firstno > lastno)
    return ("wrong stations range");
  if (firstno != stno + 1)
    {
      char errs[BUFSIZ];
      sprintf (errs, "station %d expected", stno+1);
      return (savestr (errs));
    }

  /* Allocate stations. */
  nbytes = (lastno + 1) * sizeof (station);
  if (stno == 0)
    st_table = malloc (nbytes);
  else
    st_table = realloc (st_table, nbytes);
  if (st_table == NULL)
    no_memory (nbytes);
  if (stno == 0)
    memset ((char *)&st_table[0], 0, sizeof (station));
  memset ((char *)&st_table[firstno], 0, (lastno-firstno+1) * sizeof(station));


  if (stno == 0)
    st_table[0].first_f = allocate (sizeof (frame));
  stno = lastno;
  for (j = firstno; j <= lastno; j++)
    {
      stp = &st_table[j];
      stp->first_f = allocate (cyclelen * sizeof (frame));

      /* Initialization. */
      stp->number = stp - st_table;
    }


  /* Read generators and other parameters. */
  while ((len = readline (infile, line, BUFSIZ)) >= 0)
    {
      if (len == 0)
	continue;
      lp = line;
      if (tokenis ("generator"))
	errstr = read_generator (lp, firstno, lastno);
      else if (tokenis ("streamreq"))
	errstr = read_stream_request (lp, firstno, lastno);
      else if (tokenis ("vbrreq"))
	errstr = read_vbr_request (lp, firstno, lastno);
      else if (tokenis ("maxqueuelen"))
	errstr = read_maxqueuelen (lp, firstno, lastno);
      else
	return
	  ("keyword expected at the start of line");

      if (errstr != NULL)
	return (errstr);
    } /* read generators and queue lines */

  return (NULL);
} /* read_station () */

/*
 * read_computer: read the input file and parse the compute data.
 */
static char *
read_computer (FILE * infile, char *lp)
{
  extern computer_n computers [];
  int firstno, lastno;
  computation *ctail, *clist;
  computer_n *computer_np;
  logical found;

  /* Read and check stations */
  lp = next_token (lp);

  if (lp == NULL)
    return ("station number is missing");
  if (tokenis ("sum"))
    firstno = lastno = 0;
  else if (tokenis ("all"))
    {
      firstno = 0;
      lastno = stno;
    }
  else
    {
      int noread = sscanf (lp, "%d:%d", &firstno, &lastno);
      if (noread < 1)
	return ("station number is missing");
      if (noread == 1)
	lastno = firstno;
      if (firstno < 1)
	return ("station numbers must be greater than 1");
      if (firstno > lastno)
	return ("first station number cannot be greater than the second");
      if (lastno > stno)
	return ("no stations with such a big station number");
    }

  lp = next_token (lp);
  found = FALSE;
  clist = NULL;			/* just to make the compiler quiet */
  for (computer_np = computers; computer_np->name != NULL; computer_np++)
    if (tokenis (computer_np->name))
      {
	found = TRUE;
	clist = computer_np->reader (firstno, lastno, next_token (lp));
	if (clist == NULL)
	  return (errmsg);
	break;
      }
  if (!found)
    return ("unknown computer");

  /* Add returned computer list at the tail of the computation list. */
  if (computation_list == NULL)
    computation_list = clist;
  else
    {
      for (ctail = computation_list;
	   ctail->cm_next != NULL;
	   ctail = ctail->cm_next)
	continue;
      ctail->cm_next = clist;
    }

  return (NULL);
} /* read_computer */



/*
 * read_block: read the input file and parse the data of the station and
 * computer blocks
 */
static char *
read_block (FILE * infile)
{
  char *errstr;
  char line[BUFSIZ], *lp;
  int len;

  /* Read nonblank lines */
  do
    {
      /* Skip blank and comment lines preceding a block. */
      do
	len = readline (infile, line, BUFSIZ);
      while (len < 1 && !feof (infile));

      if (feof (infile))
	return (NULL);

      lp = line;
      /* Start of a station description. */
      if (tokenis ("station") || tokenis ("stations"))
	{
	  errstr = read_station (infile, lp);
	  if (errstr != NULL)
	    return (errstr);
	}

      /* Start of a computation block */
      else if (tokenis ("compute"))
	{
	  if (stno == 0)
	    return ("no stations in input file");
	  do
	    {
	      if (len > 0)
		{
		  if (!tokenis ("compute"))
		    return ("`compute' expected");
		  errstr = read_computer (infile, lp);
		  if (errstr != NULL)
		    return (errstr);
		}
	      len = readline (infile, line, BUFSIZ);
	    }
	  while (len >= 0);
	} /* read computer lines */

      /* End block. */
      else if (tokenis ("end"))
	{
	  endinput = TRUE;
	  return (NULL);
	}

      else
	return ("`station', or `compute' or `end' expected");
    }
  while (len == 0 && !feof (infile));

  return (NULL);
} /* read_block () */

static void
print_emulate_banner ()
{
  int j;
  observable_name *onp;

  for (j = 1; j <= stno; j++)
    for (onp = observable_names; onp->name != NULL; onp++)
      switch (onp->what)
	{
	case wrong_o:
	  internal_error ();
	case s_delay_o: case v_delay_o: case i_delay_o: case b_delay_o:
	case s_trudelay_o: case v_trudelay_o: case i_trudelay_o: case b_trudelay_o:
	  break;
	default:
	  printf ("%02d%5.5s\t", j, onp->name);
	  break;
	}
  for (onp = observable_names; onp->name != NULL; onp++)
    switch (onp->what)
      {
      case wrong_o:
	internal_error ();
      case s_delay_o: case v_delay_o: case i_delay_o: case b_delay_o:
      case s_trudelay_o: case v_trudelay_o: case i_trudelay_o: case b_trudelay_o:
	break;
      default:
	printf ("++%5.5s\t", onp->name);
	break;
      }
  printf ("\n");
} /* print_emulate_banner() */


static void
register_emulation_data_for_this_frame (void)
{
  observable_name *onp;
  double val;
  station *stp;
  collection *cp;
  history *hp;


  /* If in emulation-only mode, print everything.
     Else, collect statistics and compile history. */
  if (emulateonly)
    {
      /* Print all the observable values for each station. */
      for_all_stations
	for (onp = observable_names; onp->name != NULL; onp++)
	  switch (onp->what)
	    {
	    case wrong_o:
	      internal_error ();
	    case s_delay_o: case v_delay_o: case i_delay_o: case b_delay_o:
	    case s_trudelay_o: case v_trudelay_o: case i_trudelay_o: case b_trudelay_o:
	      break;
	    default:
	      val = observable_value (stp->curr_f, onp->what);
	      printf ("%g\t", val);
	      break;
	    } /* print all observable values */

      /* Now print the sums. */
      for (onp = observable_names; onp->name != NULL; onp++)
	switch (onp->what)
	  {
	  case wrong_o:
	    internal_error ();
	  case s_delay_o: case v_delay_o: case i_delay_o: case b_delay_o:
	  case s_trudelay_o: case v_trudelay_o: case i_trudelay_o: case b_trudelay_o:
	    break;
	  default:
	    val = 0;
	    for_all_stations
	      val += observable_value (stp->curr_f, onp->what);
	    printf ("%g\t", val);
	    break;
	  } /* print all observable value sums */
      printf ("\n");
    }

  else
    {
      /* Collect the statistics for each station. */
      for_all_stations
	for (cp = stp->collection_list; cp != NULL; cp = cp->c_next)
	  switch (cp->c_what)
	    {
	      /*
	       * st_collect's second parameter is declared as
	       * double, so all second arguments are converted
	       * to double.  This is not true if a not-ANSI
	       * compiler is used, so you'll need a cast if you
	       * want to convert this code to non-ANSI.
	       */
	    case wrong_o:
	      internal_error ();
	    case s_delay_o: case v_delay_o: case i_delay_o: case b_delay_o:
	    case s_trudelay_o: case v_trudelay_o: case i_trudelay_o: case b_trudelay_o:
	      break;
	    default:
	      val = observable_value (stp->curr_f, cp->c_what);
	      st_collect (&cp->c_data, val, 1);
	      break;
	    } /* for all station's collections */

      /* Collect the statistics of the sums. */
      for (cp = st_table[0].collection_list; cp != NULL; cp = cp->c_next)
	switch (cp->c_what)
	  {
	    /*
	     * st_collect's second parameter is declared as
	     * double, so all second arguments are converted
	     * to double.  This is not true if a not-ANSI
	     * compiler is used, so you'll need a cast if you
	     * want to convert this code to non-ANSI.
	     */
	  case wrong_o:
	    internal_error ();
	  case s_delay_o: case v_delay_o: case i_delay_o: case b_delay_o:
	  case s_trudelay_o: case v_trudelay_o: case i_trudelay_o: case b_trudelay_o:
	    break;
	  default:
	    val = 0;
	    for_all_stations
	      val += observable_value (stp->curr_f, cp->c_what);
	    st_collect (&cp->c_data, val, 1);
	    break;
	  } /* collections of the sums */

      /* Compile the histories for each station. */
      for_all_stations
	for (hp = stp->history_list; hp != NULL; hp = hp->h_next)
	  if (hp->h_current <= hp->h_last)
	    switch (hp->h_what)
	      {
	      case wrong_o:
		internal_error ();
	      case s_delay_o: case v_delay_o: case i_delay_o: case b_delay_o:
	      case s_trudelay_o: case v_trudelay_o: case i_trudelay_o: case b_trudelay_o:
		break;
	      default:
		val = observable_value (stp->curr_f, hp->h_what);
		(hp->h_current++)->f = val;
		break;
	      } /* for all station's  histories */

      /* Compile the histories of the sums. */
      for (hp = st_table[0].history_list; hp != NULL; hp = hp->h_next)
	if (hp->h_current <= hp->h_last)
	  switch (hp->h_what)
	    {
	    case wrong_o:
	      internal_error ();
	    case s_delay_o: case v_delay_o: case i_delay_o: case b_delay_o:
	    case s_trudelay_o: case v_trudelay_o: case i_trudelay_o: case b_trudelay_o:
	      break;
	    default:
	      val = 0;
	      for_all_stations
		val += observable_value (stp->curr_f, hp->h_what);
	      (hp->h_current++)->f = val;
	      break;
	    } /* for all station's  histories */

    } /* if not emulate-only mode */
} /* register_emulation_data_for_this_frame() */



#define compute_queue_dropped_sent(TT) do				\
{									\
  int space = (cf->allocation.TT + cf->extraspace.TT);			\
  cf->queue.TT = pf->queue.TT + cf->input.TT - space;			\
  if (cf->queue.TT <= 0)						\
    {									\
      cf->dropped.TT = 0;						\
      cf->queue.TT = 0;							\
      cf->sent.TT = pf->queue.TT + cf->input.TT;			\
    }									\
  else									\
    {									\
      int maxqueuelen = stp->maxqueuelen.TT;				\
      if (maxqueuelen > 0 && cf->queue.TT > maxqueuelen)		\
	{								\
	  cf->dropped.TT = cf->queue.TT - maxqueuelen;			\
	  cf->queue.TT = maxqueuelen;					\
	  if (warmed)							\
	    {								\
	      stp->trus_dropped.TT = TRUE;				\
	      st_table[0].trus_dropped.TT = TRUE;			\
	    }								\
	}								\
      else								\
	{								\
	  cf->dropped.TT = 0;						\
	}								\
      cf->sent.TT = space;						\
    }									\
} while (0)

/*
 * This is the main routine.  It processes one input file at once.
 */
static void
emulate (FILE *infile)
{
  char *errstr;
  int str_total;
  tg *tgp;
  station *stp;
  computation *cmpp;
  struct tms start_cpu, stop_cpu;
  int start_real, stop_real;


  /* Clean global variables. */
  stno = 0;
  frametime = 0;
  framesize = 0;
  rttime = 0;
  histlen = 0;
  trudhistlen = 0;
  max_cslen = 0;
  cyclelen = 0;
  warm_up_time = 0;
  global_seed = 0;
  reference_traffic = 0;

  st_table = NULL;
  computation_list = NULL;
  init_station = NULL;
  compute_request = NULL;
  compute_allocation = NULL;
  stop_condition = NULL;

  totalloc = 0;
  start_real = 0;		/* to make the compiler quiet */

  if (verbose)
    {
      start_real = times (&start_cpu);
      fprintf (stderr, "Reading input file %s...\n", infilename);
    }

  /* Read global variables and set the correct readers for initers,
     requesters, allocators and stoppers. */
  if ((errstr = read_global_vars (infile)) != NULL)
    input_error (errstr, 3);
  if ((errstr = read_initers (infile)) != NULL)
    input_error (errstr, 3);
  if ((errstr = read_requesters (infile)) != NULL)
    input_error (errstr, 3);
  if ((errstr = read_allocators (infile)) != NULL)
    input_error (errstr, 3);
  if ((errstr = read_stoppers (infile)) != NULL)
    input_error (errstr, 3);

  /* Read station blocks and compute blocks. */
  endinput = FALSE;
  while (!feof (infile) && !endinput)
    {
      errstr = read_block (infile);
      if (errstr != NULL)
	input_error (errstr, 3);
    }

  /* Check the number of stations and computers. */
  if (stno == 0)
    input_error ("no stations in input file", 3);
  if (computation_list == NULL && !emulateonly)
    input_warning ("no computers in input file");

  /* Check the capacity of the frame. */
  str_total = 0;
  for_all_stations
    str_total += stp->sreq + stp->vmaxreq;
  if (str_total >= framesize)
    input_error ("sum of `sreq' and `vmaxreq' exceeds framesize", 3);

  /* Set initialisers, requesters, allocators and stoppers. */
  if ((init_station = initer_reader (initer_line)) == NULL)
    { lineno = initer_lineno; input_error (errmsg, 3); }
  if ((compute_request = requester_reader (requester_line)) == NULL)
    { lineno = requester_lineno; input_error (errmsg, 3); }
  if ((compute_allocation = allocator_reader (allocator_line)) == NULL)
    { lineno = allocator_lineno; input_error (errmsg, 3); }
  if ((stop_condition = stopper_reader (stopper_line)) == NULL)
    { lineno = stopper_lineno; input_error (errmsg, 3); }

  /* Initialise the stations. */
  stp = st_table;
  for_all_stations
    {
      stp->last_f = &stp->first_f[cyclelen - 1];
      stp->curr_f = stp->first_f;
      stp->prev_f = stp->last_f;
      stp->all_f = &stp->first_f[all_delay];
      init_station (stp);
    }

  if (verbose)
    fprintf (stderr, "  allocated %ld Kbytes -- emulating...\n",
	     (totalloc + (stno+1)*sizeof(station))/1024 + 1);

  /* Print emulate-only banner. */
  if (emulateonly && !terse)
    print_emulate_banner ();


  /* Start the time. */
  curtime = 0;
  curframeno = 1;

  /* Emulate until the stop condition is TRUE. */
  if (!checkonly) do
    {
      /* Compute frame values for each station. */
      for_all_stations
	{
	  frame *pf, *cf;

	  /* Advance frame pointers. */
	  pf = stp->prev_f = stp->curr_f;
	  stp->all_f = next_frame (stp, stp->all_f);
	  cf = stp->curr_f = next_frame (stp, stp->curr_f);

	  /* Compute things. */
	  cf->time = curtime;
	  cf->input.s = cf->input.v = cf->input.i = cf->input.b = 0;
	  for (tgp = stp->tg_list; tgp != NULL; tgp = tgp->t_next)
	    {
	      int trus = tgp->t_samplerp (&tgp->t_args);
	      switch (tgp->ttype)
		{
		case s: cf->input.s += trus; break;
		case v: cf->input.v += trus; break;
		case i: cf->input.i += trus; break;
		case b: cf->input.b += trus; break;
		}
	    }

	  /* The order stream - vbr - interactive - bulk MUST NOT be
	     changed, because that's the priority order.  Space unused
	     by stream goes to vbr, that unused by vbr goes to
	     interactive and that unused by interactive goes to bulk. */

	  compute_queue_dropped_sent (s);
	  cf->extraspace.v = cf->allocation.s - cf->sent.s;
	  compute_queue_dropped_sent (v);
	  cf->extraspace.i = cf->allocation.v + cf->extraspace.v - cf->sent.v;
	  compute_queue_dropped_sent (i);
	  cf->extraspace.b = cf->allocation.i + cf->extraspace.i - cf->sent.i;
	  compute_queue_dropped_sent (b);

	  /* Don't move this instruction from here.  Some requesters
	     (e.g. cfra) rely on it being after the computations of
	     queue and sent but before the warmup registration. */
	  cf->request = compute_request (stp);

	  /* If we are still warming up, remember the number of TRUs in
	     the queue, so one can compute the delays starting after the
	     warmup. */
	  if (!warmed)
	    stp->warmup_queue = cf->queue;
	} /* for each station */

      /* If we are warmed up then collect statistics, compile history. */
      if (warmed)
	register_emulation_data_for_this_frame ();

      compute_allocation ();
      curtime += frametime;
      if (curtime >= warm_up_time)
	warmed = TRUE;
      curframeno += 1;
    }
  while (!stop_condition ());


  /* Compute results. */
  if (!checkonly && !emulateonly)
    {
      if (verbose)
	fprintf (stderr, "  computing results...\n");

      /* Compute delay for all the stations and for the sum. */
      for_all_stations
	compute_station_delay (stp);
      compute_delay_sum ();

      /* Compute and print results. */
      if (printsrc)
	{
	  int c;

	  rewind (infile);
	  while ((c = getc(infile)) != EOF)
	    putchar (c);
	  puts ("\f");
	}
      for (cmpp = computation_list; cmpp != NULL; cmpp = cmpp->cm_next)
	cmpp->cm_computer (&st_table[cmpp->cm_st], cmpp->cm_what, &cmpp->args);
    } /* compute results if not emulateonly */


  /* End of emulation.  Output ending messages, if required. */
  if (verbose)
    {
      double realtime, cputime;

      stop_real = times(&stop_cpu);
      realtime = (stop_real - start_real) / (double)sysconf(_SC_CLK_TCK);
      cputime = (stop_cpu.tms_utime + stop_cpu.tms_stime
		 + stop_cpu.tms_cutime + stop_cpu.tms_cstime
		 - start_cpu.tms_utime - start_cpu.tms_stime
		 - start_cpu.tms_cutime - start_cpu.tms_cstime
		 ) / (double)sysconf(_SC_CLK_TCK);
      fprintf (stderr, "  done in ");
      if (realtime >= 3600)
	fprintf (stderr, "%uh%2.2um",
		 (int)(realtime/3600.0), (int)(realtime/60.0) % 60);
      else if (realtime >= 60)
	fprintf (stderr, "%um", (int)(realtime/60.0) % 60);
      fprintf (stderr, "%.2fs", fmod (realtime, 60.0));
      fprintf (stderr, " (");
      if (cputime >= 3600)
	fprintf (stderr, "%uh%2.2um",
		 (int)(cputime/3600.0), (int)(cputime/60.0) % 60);
      else if (cputime >= 60)
      fprintf (stderr, "%um", (int)(cputime/60.0) % 60);
      fprintf (stderr, "%.2fs", fmod (cputime, 60.0));
      fprintf (stderr, " cpu time).\n");
    }

  /* Reset global random state.  This is needed before freeing the
     last state used, because setstate mangles it. */
  setstate (&global_rs);

  /* Free allocated memory. */
  {
    computation *nextcmpp;
    history *nexthp;
    collection *nextcp;
    tg *nextgp;
    collection *cp;
    history *hp;
    extern void free_requesters_structures ();
    extern void free_allocators_structures ();
    extern void free_computers_structures ();

    free_computers_structures ();
    free_allocators_structures ();
    free_requesters_structures ();

    for (cmpp = computation_list; cmpp != NULL; cmpp = nextcmpp)
      {
	nextcmpp = cmpp->cm_next;
	free (cmpp);
      }

    for (stp = st_table; stp <= st_table+stno; stp++)
      {
        for (hp = stp->history_list; hp != NULL; hp = nexthp)
	  {
	    nexthp = hp->h_next;
	    free (hp);
	  }
        for (cp = stp->collection_list; cp != NULL; cp = nextcp)
	  {
	    nextcp = cp->c_next;
	    free (cp);
	  }
        for (tgp = stp->tg_list; tgp != NULL; tgp = nextgp)
	  {
	    nextgp = tgp->t_next;
	    free (tgp);
	  }
	free (stp->first_f);
      }
    free (st_table);

    if (stopper_line != NULL) free (stopper_line);
    if (allocator_line != NULL) free (allocator_line);
    if (requester_line != NULL) free (requester_line);
    if (initer_line != NULL) free (initer_line);
  }
} /* emulate() */



int
main (int argc, char **argv)
{
  extern char *optarg;
  extern int optind;
  int par;
  int opt;

  /* Store program name pointer. */
  progname = argv[0];

  /* Verify architecture.
     The union history_item members must all have the same size,
     the trans member must have the same size of an argtype_value union. */
  {
    history_item hitest;
    union argtype_value avtest;

    ASSERT (sizeof (hitest.f) == sizeof (hitest.ds));
    ASSERT (sizeof (avtest.trans) == sizeof (avtest));
  }

  /* Initialise the parameter lengths to a negative number. */
  for (par = 0; par < SUBSTPARS; par++)
    substlen[par] = -1;

  /* Read options. */
  while ((opt = getopt (argc, argv, "1:2:3:4:5:6:7:8:9:0:cestvw")) != -1)
    switch (opt)
      {
      case '1': case '2': case '3': case '4': case '5':
      case '6': case '7': case '8': case '9': case '0':
        substpar[opt-'0'] = optarg;
	substlen[opt-'0'] = strlen (optarg);
	break;
      case 'c':
	checkonly = TRUE;
	break;
      case 'e':
	emulateonly = TRUE;
	break;
      case 's':
	printsrc = TRUE;
	break;
      case 't':
	terse = TRUE;
	break;
      case 'v':
	verbose = TRUE;
	break;
      case 'w':
	kill_warnings = TRUE;
	break;
      case '?':
	usage ();
	exit (1);
	optarg = NULL;		/* use it, just to make the compiler quiet */
      }

  if (optind == argc)
    {
      usage ();
      exit (1);
    }

  for (; optind < argc; optind++)
    {
      FILE *infile;

      /* Allocation and initialisation. */
      infilename = argv[optind];
      if (streq (infilename, "-"))
	{
	  infilename = "standard input";
	  infile = stdin;
	}
      else
	infile = fopen (infilename, "r");
      if (infile == NULL)
	{
	  fprintf (stderr,
		   "%s: cannot open %s for reading\n", progname, infilename);
	  exit (2);
	}
      lineno = 0;
      do
	emulate (infile);
      while (!feof(infile));
    }

  return (0);
} /* main */
