/****************************************************************************

  $Source: /home/foda/fracas/RCS/fracas.h $
  $Revision: 7.18 $
  $Date: 1999/10/28 11:21:52 $
  $State: Exp $
  $Locker:  $

****************************************************************************/
#ident "$Id: fracas.h 7.18 1999/10/28 11:21:52 pot Exp $"

/*
 *  FRACAS stands for FRAmed Channel Access Simulator.
 */

#ifdef __GNUC__
 #include <float.h>
#endif
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/wait.h>
#include <unistd.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include <time.h>
#include "random.h"
#include "stcollect.h"

#define TRUE	1
#define FALSE	0
#define LAST	0
#define loop while(1)
typedef int logical;

#define NO_REQUEST -1

/* The frame is a struct containing the numbers relevant to a frame.
   A TRU is a traffic unit. Input traffic, output traffic and queues
     are all multiple of this basic quantity.
   BOF is beginning of frame.
   EOF is end of frame.
   Traffics are stream, vbr, interactive, bulk.
 */
typedef enum tt { s, v, i, b } tt;
typedef struct { int s, v, i, b; } itraffic;
typedef struct { float s, v, i, b; } ftraffic;
typedef struct { logical s, v, i, b; } ltraffic;

typedef struct
{
  int time;			/* starting time of this frame */
  itraffic allocation;		/* TRUs that can be sent at EOF */
  itraffic extraspace;		/* TRUs from higher priority traffic types */
  itraffic input;		/* TRUs on input during the frame */
  itraffic queue;		/* TRUs in the input queue at EOF */
  itraffic dropped;		/* input TRUs that exceeded the queue length */
  itraffic sent;		/* TRUs sent at EOF */
  itraffic request;		/* TRUs requested at EOF */
} frame;


/*
 *  The tg (traffic generator) is a struct containing the status
 *  information of a generator.  tg's are kept in linked lists
 *  whose head is contained in the station struct.
 *  A sampler is a function that takes a tg_args union as argument
 *  and returns the number of TRUs generated in the current frame.
 */
struct cycle_vars
{
  float start;			/* cycle start time */
  float stop;			/* cycle stop time (0 means infinite) */
  float cycle;			/* cycle length (0 means infinite) */
  logical on;			/* generator is transmitting */
  float next_on_switch;		/* next time we will switch on */
  float next_off_switch;	/* next time we will switch off */
};

typedef union
{
  struct constant_args
    {
      unsigned int burst;	/* TRUs in a burst */
      float interval;		/* time interval of TRUs */
      float last_gen;		/* time of last TRU generated */
      struct cycle_vars cv;	/* cycle handling data */
    } constant;			/* args to constant sampler */
  struct poisson_args
    {
      random_state rs;
      unsigned int burst;	/* TRUs in a burst */
      float lambda;		/* average # of TRUs per frame */
      double LambdaExp;		/* exp of mean # of TRUs per frame */
      struct cycle_vars cv;	/* cycle handling data */
    } poisson;			/* args to poisson sampler */
  struct impulse_args
    {
      random_state rs;
      unsigned int burst;	/* TRUs in a burst */
      float interval1, interval2; /* mean intv. between TRUs in each state */
      double Lambda1Exp, Lambda2Exp; /* exp of mean # of TRUs per frame */
      float u11, u22;		/* mean time of permanence in each state */
      int state;		/* current state */
    } impulse;			/* args to 2-state MMPP sampler */
  struct markov_args
    {
      random_state rs;
      unsigned int burst;       /* TRUs in a burst */
      float binsize;		/* width of the relative throughput bins */
      float steplen;		/* lenght of the time step of the chain (TRU) */
      int stepno;		/* # of steps after which the chain rewinds */
      float next_switch;	/* time to next state change */
      int cur_step;		/* # of steps done so far */
      struct ma_state {		/* --- list element: a graph node --- */
	struct ma_state *nextstate;
	int state;		/* state number */
	float iniprob;		/* probability of being initial state */
	float thput;		/* normalised throughput for this state */
	struct ma_edge {	/* -- sublist element: a graph edge -- */
	  struct ma_edge *nextedge;
	  int st;		/* # of destination state of the edge */
	  struct ma_state *stp;	/* ptr. to destination state of the edge */
	  double cumjmpprob;	/* cumulative prob. to switch to dest. state */
	} *ehead;		/* head of the list of edges */
      } *chain;			/* head of the list of states */
      struct ma_state *cur_sp;	/* pointer to the current state */
      double cur_thput;		/* throughput of the current state */
    } markov;			/* args to markov sampler */
 struct enst_bulk_args
    {
      random_state rs;
      unsigned int burst;	/* TRUs in a burst */
      unsigned int backlog;	/* TRUs generated by files and not yet sent */
      float intervalfiletime;	/* mean interval between file generations */
      double LambdaFiletimeExp; /* exp of mean # of files per frame */
      float fileminlen, filelen; /* minimum and mean exp file length */
      double invfilelenExp;	/* exp of mean of inverse of exp file length */
      float constinterval;	/* interval between TRUs in state 1 */
      float mininterval;	/* min interval between TRUs in state 2 */
      float interval2;		/* mean of exp intv. between TRUs in state 2 */
      float u11, u22;		/* mean time of permanence in each state */
      int state;		/* current state */
    } enst_bulk;		/* args to ENST bulk sampler */
  struct enst_interactive_args
    {
      random_state rs;
      unsigned int burst;	/* TRUs in a burst */
      float mininterval;	/* min interval between packets */
      float expinterval;	/* exponential interval between packets */
      float firstinterval;	/* interval from start of frame to first pkt */
      struct {
	float cumdistr;
	int packet_len;
      } *len_distrib;		/* packet length distribution */
      int pld_bins;		/* length of the len_distrib array */
    } enst_interactive;		/* args to ENST interactive sampler */
  struct fgn_args
    {
      random_state rs;
      unsigned int burst;	/* TRUs in a burst */
      float average;		/* average # of TRUs per TMU */
      float peakedness;		/* peakedness of # of TRUs per TMU */
      float H;			/* Hurst parameter */
      int rmdn;			/* base-2 log of the length of the trace */
      float *firstp;		/* pointer to first sample */
      float *lastp;		/* pointer to last sample */
      float *samplep;		/* pointer to current sample */
    } fgn;
  struct file_args
    {
      unsigned int burst;	/* TRUs in a burst */
      FILE *stream;		/* list of TRUs or intervals */
      logical interval;		/* interval or packet mode */
      float last_time;		/* time of last TRU generated */
      float scale;		/* factor to divide the intervals */
    } file;
} tg_args;

typedef int sampler (tg_args *argsp);

typedef struct tg
{
  struct tg *t_next;		/* pointer to the next in linked lists */
  tt ttype;			/* type of traffic */
  sampler *t_samplerp;		/* pointer to a sampler function */
  tg_args t_args;		/* arguments to the sampler */
} tg;

typedef tg *sampler_r (char *argline);
typedef struct sampler_n { sampler_r *reader; char *name; } sampler_n;

/*
 * The measurable quantities are listed in the observable enum.
 */

typedef enum observable
{
  wrong_o,				/* nonexistent identifier */
  /* TRUs entering a station */
  s_input_o, v_input_o, i_input_o, b_input_o,
  /* TRUs in queue */
  s_queue_o, v_queue_o, i_queue_o, b_queue_o,
  /* TRUs in queue plus TRUs entering */
  s_inbuffer_o, v_inbuffer_o, i_inbuffer_o, b_inbuffer_o,
  /* TRUs requested */
  s_request_o, v_request_o, i_request_o, b_request_o,
  /* TRUs allocated */
  s_alloc_o, v_alloc_o, i_alloc_o, b_alloc_o,
  /* TRUs sent */
  s_sent_o, v_sent_o, i_sent_o, b_sent_o,
  /* extraspace of TRUs */
  v_extra_o, i_extra_o,	 b_extra_o,
  /* unused allocation */
  b_unused_o,
  /* delay of entering TRUs (per frame)*/
  s_delay_o, v_delay_o, i_delay_o, b_delay_o,
  /* delay of entering TRUs */
  s_trudelay_o, v_trudelay_o, i_trudelay_o, b_trudelay_o
} observable;

typedef struct observable_name
{
  observable what;
  char *name;
} observable_name;

extern observable_name observable_names [];

typedef struct collection
{
  struct collection *c_next;	/* pointer to the next element in list */
  observable c_what;		/* observable to collect */
  st_data c_data;		/* collected data (see stcollect.h) */
} collection;

typedef struct
{
  unsigned short n, val;
} dshort;

typedef union
{
  float f;			/* float history item */
  dshort ds;			/* double short history item */
} history_item;

typedef struct history
{
  struct history *h_next;	/* pointer to the next element in list */
  observable h_what;		/* observable to register in history */
  history_item *h_last;		/* pointer to the last history item */
  history_item *h_current;	/* pointer to the first empty h. item */
  history_item h_items[1];	/* first history item */
				/* malloc'ed array of history items */
} history;

typedef union
{
  struct cfra_rd
    {
      enum { noburst, burst } state;
      int burstlen;
      int silentframes;
    } cfra;
  struct fodaibea_rd
    {
      int cum_dtraffic;
      int cum_dframes;
      int last_cum_dtraffic;
      int cum_vtraffic;
      int *vtraffics;		/* array of traffic */
      double step;		/* step between vbr request levels */
      int curlev;		/* currently requested level */
      int rqueue;		/* vbr real queue */
      int vqueue;		/* vbr virtual queue */
    } fodaibea;
} reqdata;

/*
 *  The station is a struct containing all the information for a station.
 */

typedef struct
{
  long int number;		/* station number */
  itraffic maxqueuelen;		/* max input queue lengths */
  ltraffic trus_dropped;	/* some TRU dropped because of queue length */
  itraffic warmup_queue;	/* queue when warmup ends */
  int sreq;			/* stream request */
  int vminreq, vmaxreq;		/* min and max vbr requests */
  float Ji;			/* input jitter for vbr */
  frame *first_f;		/* pointer to first frame (array) */
  frame *last_f;		/* pointer to last frame (array) */
  frame *prev_f;		/* pointer to previous frame (ring) */
  frame *curr_f;		/* pointer to current frame (ring) */
  frame *all_f;			/* pointer to allocation frame (ring) */
  tg *tg_list;			/* head of list of tgs */
  collection *collection_list;	/* head of a list of collections */
  history *history_list;	/* head of a list of histories  */
  reqdata rd;			/* requester data */
} station;

#define previous_frame(stationp, framep) \
  (((framep) == (stationp)->first_f) ? (stationp)->last_f : (framep)-1)
#define next_frame(stationp, framep) \
  (((framep) == (stationp)->last_f) ? (stationp)->first_f : (framep)+1)

/*
 * The following is what is needed to list the computations to do and
 * print the results.
 */

typedef union
{
  struct
    {
      FILE *ostream;
    } simplestats;
  struct
    {
      logical aggregate, compact;
      int every;
      FILE *ostream;
    } listing;
  struct
    {
      float q;
      FILE *ostream;
    } quantile;
  struct
    {
      int catn;
      FILE *ostream;
    } categorise;
  struct
    {
      int rows, cols;
      FILE *ostream;
    } ttydistrib;
  struct
    {
      int rows, cols;
      FILE *ostream;
    } ttygraph;
} cm_args;

typedef void computer (station *, observable, cm_args *);

typedef struct computation
{
  struct computation *cm_next;	/* pointer to the next element in list */
  int cm_st;			/* station index */
  observable cm_what;		/* observable to make computations on */
  computer *cm_computer;	/* computer for making computations */
  cm_args args;			/* arguments to the computer */
} computation;

typedef computation *computer_r (int firstno, int lastno, char *argline);
typedef struct computer_n { computer_r *reader; char *name; } computer_n;

/*
 * This structure is used for defining the arguments expected by a reader.
 */
typedef struct
{
  char *name;
  enum argtype_type
    { int_t, uns_t, float_t, string_t, log_t,
      ifile_t, ofile_t, afile_t, ipipe_t, opipe_t, iopipe_t
    } type;
  enum { required, optional, got } state;
  #define LONGEST_TYPE unsigned long long int
  union argtype_value
    {
      int int_v;
      unsigned int uns_v;
      float float_v;
      char *string_v;
      logical log_v;
      FILE *stream_v;
      LONGEST_TYPE trans;
    } value;
} arg_type;

/*
 * This macro is used to get the value of an entry in an array of arg_type's.
 */
#define get_arg(name, var)  _get_value (args, (name), &(var));

/*
 * This macro is used by readers to return a constant string error.
 */
#define return_error(s) do { \
 strncpy (errmsg, (s), BUFSIZ); errmsg[BUFSIZ-1] = '\0'; return 0; } while (0)

/*
 * This macro is used to loop through all the station pointers.
 */
#define for_all_stations for (stp=st_table+1; stp<=st_table+stno; stp++)

/*
 * Typedefs for functions.
 */
typedef void initer (station *stp);
typedef initer *initer_r (char *argline);
typedef struct initer_n { initer_r *reader; char *name; } initer_n;
typedef itraffic requester (station *stp);
typedef requester *requester_r (char *argline);
typedef struct requester_n { requester_r *reader; char *name; }requester_n;
typedef void allocator (void);
typedef allocator *allocator_r (char *argline);
typedef struct allocator_n { allocator_r *reader; char *name; }allocator_n;
typedef logical stopper (void);
typedef stopper *stopper_r (char *argline);
typedef struct stopper_n { stopper_r *reader; char *name; } stopper_n;

/*
 * Global variables initialised from input file.
 */
extern int stno;		/* number of stations */
extern int frametime;		/* duration of a frame */
extern int framesize;		/* capacity of the frame (no. of TRUs) */
extern int rttime;		/* round trip time */
extern int histlen;		/* lenght of histories */
extern int trudhistlen;		/* TRU delay history length */
extern int max_cslen;		/* max length used for catn=0 */
extern int cyclelen;		/* length of the big table (frames) */
extern int warm_up_time;	/* initial simulation time without stats */
extern float reference_traffic;	/* global reference value for traffic */

/*
 * Internal global variables.
 */
extern logical terse;		/* don't print headers on stdout */
extern char errmsg[BUFSIZ];	/* error messages buffer */
extern int lineno;		/* current input file line number */
extern station *st_table;	/* the stations array */
extern int curtime;		/* BOF of the current frame */
extern int curframeno;		/* frame number, starting from 1 */
extern int all_delay;		/* # of frames from curr_f to all_f */
extern long int totalloc;	/* total bytes allocated */
extern logical warmed;		/* curtime >= warm_up_time */
extern computation *computation_list, *computation_tail;

/*
 * Global macros.
 */
/* string is the first token in the string pointed by lp */
#define tokenis(token) first_token_eq (lp, token)
#define SEPS " \t"
#define TOKS "abcdefghijklmnopqrstuvwzyzABCDEFGHIJKLMNOPQRSTUVWZYZ0123456789-_"
#define streq(s1,s2) (strcmp (s1, s2) == 0)
#define excess_intdiv(n,d) ((n) % (d) == 0 ? (n) / (d) : (n) / (d) + 1)
#define internal_error() _internal_error (__FILE__, __LINE__)
#define ASSERT(condition) do if (!(condition)) internal_error (); while (0)
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

/*
 * Global functions mostly defined in fracas.c.
 */
extern logical first_token_eq (char *line, char *tok);
extern void input_warning (char *message);
extern void runtime_error (char *message);
extern void no_memory (int n);
extern void _internal_error (char *file, int line);
#ifndef DEBUG
 extern void *allocate (int);	/* allocate memory with check */
#endif
extern char *observable_string (observable what);
extern char *concat3 (char *, char *, char *);
extern char *savestr (char *), *savenstr (char *, int);
extern char *lineparser (char *lp, arg_type *args);
extern char *lineparser_with_ostream (char *lp, arg_type *args, FILE **outstr);
extern char *next_token (char *lp);
extern void _get_value (arg_type *args, char *name, void *varp);
extern void close_and_wait (FILE *stream);
extern void compute_station_delay (station *), compute_delay_sum (void);


/*
 * Prototypes of library functions.
 */
extern int errno;
extern const char *const sys_errlist[];
#ifdef DEBUG
 #include "chkmalloc.h"
 #define allocate(n) (totalloc += n, trace_calloc(__FILE__, __LINE__, n, 1))
#endif /* DEBUG */
