/* Returns a normally distributed deviate with zero mean and unit
   variance, using the int randgen(void) uniform random generator,
   which is supposed to return a random number in the range [0;INT_MAX]. */

#include <limits.h>
#include <math.h>

float
gasdev (randgen)
     int randgen ();
{
  static int iset = 0;
  static float gset;
  static float s = 2.0/((float)INT_MAX+1);
  float fac, rsq, v1, v2;

  if (iset)
    {				/* We have an extra deviate handy, */
      iset = 0;			/* so unset the flag, */
      return (gset);		/* and return it. */
    }
    
  do				/* We don't have an extra deviate handy, */
    {				/* so pick two uniform numbers in the */
      v1 = s*randgen()-1;	/* square extending from -1 to +1 in each */
      v2 = s*randgen()-1;	/* direction, see if they are in the unit */
      rsq = v1*v1 + v2*v2;	/* circle, and if they are not, try again. */
    }
  while (rsq >= 1 || rsq == 0);

  /* Now make the Box-Muller transformation to get two normal
     deviates.  Return  one and save the other for next time. */
  fac = sqrt(-2.0*log(rsq)/rsq);
  gset = v1 * fac;
  iset = 1;			/* Set flag. */
  return (v2 * fac);
}
