/* Both Motorola Delta's and Digital Alpha popen/pclose are broken,
   so here's my own version.  It does not block the signals around
   the wait(2).  Should it? */
#include <sys/types.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>

#define	READER	0
#define	WRITER	1

#if (!defined (MAX_FOPEN))
  #if (defined (MAX_OPEN))
    #define MAX_FOPEN MAX_OPEN
  #elif (defined (_NFILE))
    #define MAX_FOPEN _NFILE
  #else
    #define MAX_FOPEN 1000
  #endif
#endif

extern FILE *fdopen(int fd, const char *type);
extern int execl(const char *path, const char *arg, ...);
extern pid_t fork(void);
extern int pipe(int fd[2]);
extern int close(int fd);
#ifdef m68k
extern int fcntl(int fd, int cmd, int arg);
#endif
extern int wait(int *stat_loc);

static struct { FILE *str; pid_t pid; } pipes [MAX_FOPEN];

FILE *
pot_popen (char *command, char *type)
{
  int pipends [2];
  register int fno, parentend, childend;
  register pid_t pid;

  if (pipe (pipends) < 0)
    return (NULL);
  if (type[0] == 'r')
    {
      parentend = pipends[READER];
      childend = pipends[WRITER];
    }
  else
    {
      parentend = pipends[WRITER];
      childend = pipends[READER];
    }
  pid = fork ();

  /* The child side exits with execl and never returns. */
  if (pid == 0)
    {
      /* Close all pipes from other pot_popen's. */
      for (fno = 0; fno < MAX_FOPEN; fno++)
	if (pipes[fno].pid != 0)
	  fclose (pipes[fno].str);

      close (parentend);
      if (type[0] == 'r')
	{
	  fclose (stdout);
	  fcntl (childend, F_DUPFD, 1);
	}
      else
	{
	  fclose (stdin);
	  fcntl (childend, F_DUPFD, 0);
	}
      close (childend);
      execl ("/bin/sh", "sh", "-c", command, NULL);
    }

  /* This is the parent side. */
  if (pid == -1)
    return (NULL);
  close (childend);

  pipes[parentend].pid = pid;
  pipes[parentend].str = fdopen (parentend, type);

  return (pipes[parentend].str);
}

int
pot_pclose (FILE *stream)
{
  register int child, fno;
  int filno, status;

  filno = fileno (stream);
  status = fclose (stream);

  if (pipes[filno].pid == 0)
    {
      if (status == 0)
	errno = ECHILD;
      status = -1;
    }
  else
    {
      fno = -1;
      do
	{
	  child = wait (&status);
	  if (child == -1)
	    {
	      if (errno == EINTR)
		continue;
	      status = -1;
	      break;
	    }
	  for (fno = 0; fno < MAX_FOPEN; fno++)
	    if (child == pipes[fno].pid)
	      {
		pipes[fno].pid = 0;
		if (fno != filno)
		  fclose (pipes[fno].str);
		break;
	      }
	}
      while (filno != fno);
    }

  return (status);
}
