/****************************************************************************

  $Source: /home/foda/fracas/RCS/requesters.c $
  $Revision: 7.19 $
  $Date: 1999/01/21 14:28:06 $
  $State: Exp $
  $Locker:  $

****************************************************************************/
#ident "$Id: requesters.c 7.19 1999/01/21 14:28:06 pot Exp $"

/*
 *  FRACAS stands for FRAmed Channel Access Simulator.
 */

#include "fracas.h"

static requester_r queue_reader;
static requester_r cfra_reader;
static requester_r fodaibea_reader;
static requester_r feeders_drifs_reader;

requester_n requesters [] =
{
  { queue_reader, "queue" },
  { cfra_reader, "cfra" },
  { fodaibea_reader, "fodaibea" },
  { feeders_drifs_reader, "feeders-drifs" },
  { NULL }
};

/* tracing */
#define TRACE0(f) do if (trace && warmed) printf (f); while (0)
#define TRACE1(f,a) do if (trace && warmed) printf (f,a); while (0)
#define TRACE2(f,a,b) do if (trace && warmed) printf (f,a,b); while (0)

static logical trace;

/*
 * Queue requester.
 * Makes a request equal to the current queue length.
 *
 * Syntax is: requester queue
 */
static itraffic
queue_requester (station *stp)
{
  return (stp->curr_f->queue);
} /* queue() */

static requester *
queue_reader (char *lp)
{
  return (&queue_requester);
} /* queue_reader() */

/*
 * CFRA requester.
 * Makes a request equal to sreq for stream traffic.  Makes no request
 * for the vbr traffic. Sends an `S' message at the start of an
 * incoming datagram burst, and an `E' message at its end.  The start
 * is when the incoming burts length exceeds L0 TRUs, ends when no
 * TRUs are received for the timeout lenght's time.
 *
 * Syntax is: requester cfra ctlsize L0 [timeout]
 *	Where <L0> is the minimum length of a burst in TRUs and
 *	<timeout> is the number of frames after which a burst is
 *	declared ended (default 1).  <ctlsize> is the length of a
 *	message, which is sent in band.
 */
static struct
{
  unsigned int L0;		/* threshold for sending an `S' message */
  unsigned int timeout;		/* threshold for sending an `E' message */
  unsigned int ctlsize;		/* size in TRUs of a control message */
} cfra;

static itraffic
cfra_requester (station *stp)
{
  itraffic result;
  int binput, drequest;
  struct cfra_rd *rdp = &stp->rd.cfra;


  /* Compute datagram request. */
  binput = stp->curr_f->input.b;
  drequest = NO_REQUEST;
  if (binput > 0)
    {
      rdp->burstlen += binput;
      if (rdp->state == noburst && rdp->burstlen >= cfra.L0)
	{
	  drequest = 0;		/* `S' message */
	  rdp->state = burst;
	}
      rdp->silentframes = 0;
    }
  else /* binput == 0 */
    {
      rdp->silentframes += 1;
      if (rdp->state == burst && rdp->silentframes >= cfra.timeout)
	{
	  drequest = rdp->burstlen; /* `E' message */
	  rdp->state = noburst;
	}
      rdp->burstlen = 0;
    }

  /* If a message is to be sent, consider an interactive slot as
     being occupied: if they are all full, put an interactive slot
     back in the interactive queue; else don't consider it as
     extraspace for bulk.  Note that all allocation is given for
     interactive traffic: bulk traffic only goes in extraspace. */
  if (drequest != NO_REQUEST)
    {
      unsigned int ctlsize_i, ctlsize_b;
      frame *cf = stp->curr_f;

      if (cf->extraspace.b >= cfra.ctlsize)
	{
	  ctlsize_i = 0;
	  ctlsize_b = cfra.ctlsize;
	}
      else
	{
	  ctlsize_i = cfra.ctlsize - cf->extraspace.b;
	  ctlsize_b = cf->extraspace.b;
	}

      if (ctlsize_b > 0)
	/* some unused interactive space has been assigned to bulk
	   traffic as extraspace: part of the request message is put
	   into the bulk extraspace */
	{
	  int excess_b;

	  cf->extraspace.b -= ctlsize_b;
	  excess_b = cf->sent.b - cf->allocation.b - cf->extraspace.b;

	  if (excess_b > 0)
	    /* we have sent too much bulk when in fact there is no
	       space for it (since we have reduced the bulk
	       extraspace); put some bulk traffic back in the bulk
	       queue and check for the queue length */
	    {
	      int maxqueuelen = stp->maxqueuelen.b;

	      cf->sent.b -= excess_b;
	      cf->queue.b += excess_b;
	      if (maxqueuelen > 0 && cf->queue.b > maxqueuelen)
		{
		  cf->dropped.b += cf->queue.b - maxqueuelen;
		  cf->queue.b = maxqueuelen;
		  if (warmed)
		    {
		      stp->trus_dropped.b = TRUE;
		      st_table[0].trus_dropped.b = TRUE;
		    }
		}
	    }
	}

      if (ctlsize_i > 0)
	/* the bulk extraspace is not enogh to contain the control
	   message, so use some interactive space by putting some
	   traffic back in the queue, and check the queue length */
	{
	  int maxqueuelen = stp->maxqueuelen.i;

	  cf->sent.i -= ctlsize_i;
	  ASSERT (cf->sent.i >= 0);
	  cf->queue.i += ctlsize_i;
	  if (maxqueuelen > 0 && cf->queue.i > maxqueuelen)
	    {
	      cf->dropped.i += cf->queue.i - maxqueuelen;
	      cf->queue.i = maxqueuelen;
	      if (warmed)
		{
		  stp->trus_dropped.i = TRUE;
		  st_table[0].trus_dropped.i = TRUE;
		}
	    }
	}
    }

  result.s = stp->sreq;
  result.v = NO_REQUEST;
  result.i = drequest;
  result.b = NO_REQUEST;

  return (result);
} /* cfra_requester () */

static requester *
cfra_reader (char *lp)
{
  char *errstr;
  arg_type args[] =
  {
    { "L0", uns_t, required },
    { "timeout", uns_t, optional },
    { "ctlsize", uns_t, required },
    { NULL }
  };

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);

  get_arg ("L0", cfra.L0);
  get_arg ("timeout", cfra.timeout);
  get_arg ("ctlsize", cfra.ctlsize);

  if (cfra.L0 == 0)
    return_error ("L0 must be greater than 0");
  if (cfra.timeout == 0)
    cfra.timeout = 1;
  if (cfra.ctlsize == 0)
    return_error ("ctlsize must be greater than 0");

  return (&cfra_requester);
} /* cfra_reader() */

/*
 * FODA/IBEA requester.
 * Makes a datagram request equal to the queue length plus dH times the
 * input traffic, dH being a temporal constant for the datagram traffic.
 * Makes a request equal to sreq for stream traffic.  The datagram input
 * traffic estimation is changed every dwin frames, using an average over
 * dwin frames, while the input vbr traffic is a moving average over the
 * last vwin frames.  The request for vbr traffic is between vminreq and
 * vmaxreq.  There are vlevels possible values for the vbr traffic
 * equispaced between vminreq and vmaxreq as defined for eah station.  The
 * choice of asking for a higher level is made depending on the queue
 * length, the choice of asking for a lower level is made depending on the
 * input.  The value of the input jitter Ji as defined for each station is
 * taken into account as described in
 * 
 *    N. Celandroni, M. Conti, E. Ferro, E. Gregori, F. Potorti` "A
 *    bandwidth assignment algorithm on a satellite channel for VBR
 *    traffic", International Journal on Satellite Communications, Vol. 15,
 *    No. 6, pp. 237-246, November-December 1997.
 * 
 * Syntax is: requester fodaibea dH [dwin] [vwin] [vlevels] [trace]
 *	Where the optional <trace> argument is either `y' or `n' to enable
 *	or disable the feature (disabled by default).
 *
 *	Where <dH> is a non negative floating point number, <dwin> is a
 *	positive integer defaulting to 1, <vwin> is a positive integer
 *	defaulting to 1, <vlevels> is the number of possible different
 *	vbr requests, default is 2.
 */
static struct
{
  float dH;
  unsigned int dwin;
  /* float vH; */
  unsigned int vwin;
  unsigned int vlevels;
  logical always_make_request;
} fi;

static itraffic
fodaibea_requester (station *stp)
{
  itraffic result;
  int vindex;
  struct fodaibea_rd *rdp = &stp->rd.fodaibea;

  /* Datagram cumulated traffic. */
  rdp->cum_dtraffic += stp->curr_f->input.i + stp->curr_f->input.b;
  if (++rdp->cum_dframes == fi.dwin)
    {
      rdp->last_cum_dtraffic = rdp->cum_dtraffic;
      rdp->cum_dtraffic = 0;
      rdp->cum_dframes = 0;
    }

  /* VBR traffic. */
  if (fi.vwin == 1)
    rdp->cum_vtraffic = stp->curr_f->input.v;
  else
    {
      vindex = curframeno % fi.vwin;
      rdp->cum_vtraffic -= rdp->vtraffics[vindex];
      rdp->cum_vtraffic += rdp->vtraffics[vindex] = stp->curr_f->input.v;
    }

  /* Set requests. */
  result.s = stp->sreq;
  if (fi.always_make_request
      || stp->curr_f->allocation.s > 0
      || stp->curr_f->allocation.v > 0
      || stp->curr_f->allocation.i > 0
      || stp->curr_f->allocation.b > 0)

    {
      double step = rdp->step;
      int curlev = rdp->curlev;
      int curalloc = floor (curlev * step + stp->vminreq + 0.5);
      int rqueue = rdp->rqueue + stp->curr_f->input.v - curalloc;

      if (rqueue > 0)		/* going up levels */
	{	
	  if (curlev == fi.vlevels - 1)
	    TRACE0 ("*we are at the highest level: cannot grow any more\n");
	  else
	    {
	      double threshold = stp->Ji * curalloc;
	      TRACE2 ("\n*rqueue = %d, threshold = %g\n", rqueue, threshold);

	      if (rqueue > threshold)
		{
		  if ((curlev += ceil (rqueue / step)) > fi.vlevels - 1)
		    curlev = fi.vlevels - 1;
		  TRACE1 ("*up to level %d\n", curlev);
		}
	      else
		TRACE0 ("*we stay at the same level\n");
	    } /* not at the highest level */
  
	  rdp->rqueue = rqueue;
	  rdp->vqueue = 0;
	} /* rqueue > 0 */

      else			/* staying there or going down levels */
	{ 
	  int loweralloc = floor ((curlev-1) * step + stp->vminreq + 0.5);
	  int vqueue = rdp->vqueue + stp->curr_f->input.v - loweralloc;

	  if (vqueue > 0)	/* staying at the same level */
	    vqueue =  0;
	  else			/* going down levels */
	    {
	      if (curlev == 0)
		TRACE0 ("*we are at the lowest level: cannot shrink any more\n");
	      else
		{ 
		  double threshold = stp->Ji * loweralloc;
		  TRACE2 ("\n*vqueue = %d, threshold = %g\n", vqueue, threshold);

		  if (vqueue < -threshold)
		    {
		      if ((curlev -= 1 + floor (-vqueue / step)) < 0)
			curlev = 0;
		      TRACE1 ("*down to level %d\n", curlev);
		    }
		  else
		    TRACE0 ("*we stay at the same level\n");
		} /* not at the lowest level */
	    } /* vqueue < 0 */

	  rdp->rqueue = 0;
	  rdp->vqueue = vqueue;
	} /*  rqueue <= 0 */

      rdp->curlev = curlev;
      result.v = floor (curlev * step + stp->vminreq + 0.5);
      result.i = (stp->curr_f->queue.i + stp->curr_f->queue.b
		  + fi.dH * rdp->last_cum_dtraffic);
      result.b = NO_REQUEST;
    } /* compute request */

  else
    {
      result.v = 0;
      result.i = NO_REQUEST;
      result.b = NO_REQUEST;
    }

  return (result);
  
} /* fodaibea_requester () */

static requester *
fodaibea_reader (char *lp)
{
  station *stp;
  char *errstr;
  arg_type args[] =
  {
    { "dH", float_t, required },
    { "dwin", uns_t, optional },
    /* { "vH", float_t, required }, */
    { "vwin", uns_t, optional },
    { "vlevels", uns_t, optional },
    { "trace", log_t, optional },
  { NULL }
  };

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);

  get_arg ("dwin", fi.dwin);
  get_arg ("dH", fi.dH);
  get_arg ("vwin", fi.vwin);
  /* get_arg ("vH", fi.vH); */
  get_arg ("vlevels", fi.vlevels);
  get_arg ("trace", trace);

  if (fi.dwin == 0)
    fi.dwin = 1;
  if (fi.dH < 0)
    return_error ("dH is negative");
  fi.dH /= frametime * fi.dwin;
  if (fi.vwin == 0)
    fi.vwin = 1;
  if (fi.vlevels == 0)
    fi.vlevels = 2;
  else     
    if (fi.vlevels < 2) 
      return_error ("parameter `vlevels' must be greater than 1");
  #if 0
  if (fi.vH < 0)
    return_error ("vH is negative");
  fi.vH /= frametime * fi.vwin;
  #endif

  fi.always_make_request = FALSE;

  for_all_stations
    {
      if (fi.vwin > 1)
	stp->rd.fodaibea.vtraffics
	  = allocate (fi.vwin * sizeof (*stp->rd.fodaibea.vtraffics));
      stp->rd.fodaibea.curlev = 0;
      stp->rd.fodaibea.rqueue = 0;
      stp->rd.fodaibea.vqueue = 0;
      stp->rd.fodaibea.step =
	((double)stp->vmaxreq - stp->vminreq) / (fi.vlevels - 1);
     
      if (2 * stp->vmaxreq * stp->Ji >= stp->rd.fodaibea.step)
	{
	  char warn [200];
	  sprintf (warn,
		   "parameter 'Ji' is too big for station %ld", stp->number);
	  input_warning (warn);
	}
    }
      
  return (&fodaibea_requester);
} /* fodaibea_reader() */

static requester *
feeders_drifs_reader (char *lp)
{
  requester *retval = fodaibea_reader (lp);
  fi.always_make_request = TRUE;
  return (retval);
} /* feeders_drifs_reader() */

void
free_requesters_structures ()
{
  extern requester *compute_request;
  station *stp;

  if (compute_request == fodaibea_requester)
    for_all_stations
      if (stp->rd.fodaibea.vtraffics != NULL)
	free (stp->rd.fodaibea.vtraffics);
}
