/* 
 * nrselect: Returns the kth smallest value in the array arr[1..n].
 *         The input array will be rearranged to have this value
 *         in location arr[k], with all smaller elements moved to
 *         arr[1..k-1] (in arbitrary order) and all larger elements
 *         in arr[k+1..n] (also in arbitrary order).
 */
#define SWAP(a,b) (temp=(a), (a)=(b), (b)=temp)

float nrselect (unsigned long k, unsigned long n, float arr[])
{
  unsigned long i, ir, j, l, mid;
  float a, temp;

  l = 1;
  ir = n;
  for (;;)
    {
      if (ir <= l+1)	/* Active partition contains 1 or 2 elements */
	{
	  if (ir == l+1 && arr[ir] < arr[l])	/* Case of 2 elements */
	    SWAP(arr[l],arr[ir]);
	  return (arr[k]);
	}
      else
	{ /* Chose median of left, center, and right elements as
	     partitioning element a. Also rearrenge so that 
	     arr[l+1] <= arr[l], arr[ir] >= arr[l]. */
	  mid = (l + ir) >> 1;		
	  SWAP(arr[mid],arr[l+1]);
	  if (arr[l+1] > arr[ir])
	    SWAP(arr[l+1],arr[ir]);
          if (arr[l] > arr[ir])
	    SWAP(arr[l],arr[ir]);
          if (arr[l+1] > arr[l])
	    SWAP(arr[l+1],arr[l]);
          i = l+1;		/* Initialize poiners for partitioning */
	  j = ir;
	  a = arr[l];		/* Partitioning element */
	  for (;;)		/* Beginning of innermost loop */
	    {
	      do i++; 		/* Scan up to find element > a */	
		while (arr[i] < a);
	      do j--;		/* Scan down to find element < a */
		while (arr[j] > a);
	      if (j < i) break; /* Pointers crossed.Partitioning complete */
	      SWAP(arr[i],arr[j]);
	    }			/* End of innermost loop */
	  arr[l] = arr[j];	/* Insert partitioning element */
	  arr[j] = a;
	  if (j >= k) 		/* Keep active the partition that 
				   contains the kth element */
	    ir = j-1;
	  if (j <= k) 
	    l = i;
	}
     }
 } /* nrselect () */

