/****************************************************************************

  $Source: /home/foda/fracas/RCS/stoppers.c $
  $Revision: 7.3 $
  $Date: 1996/01/19 12:45:02 $
  $State: Exp $
  $Locker:  $

****************************************************************************/
#ident "$Id: stoppers.c 7.3 1996/01/19 12:45:02 pot Exp $"

/*
 *  FRACAS stands for FRAmed Channel Access Simulator.
 */

#include "fracas.h"

static stopper_r maxtime_reader;

stopper_n stoppers [] =
{
  { maxtime_reader, "maxtime" },
  { NULL }
};



/*
 * Max time stopper.
 * Stops the simulation when a given number of time units or frames
 * has elapsed.
 *
 * Syntax is: stopper maxtime <time> <frames>
 *	where <time> (<frames>) is the number of time units (frames)
 *	after which the simulation must be stopped.  Only one of <time>
 *	or <frames> is allowed.
 */
static struct
{
  int time;
} maxtime;

static logical
maxtime_stopper (void)
{
  return (curtime >= maxtime.time);
} /* maxtime_stopper() */

static stopper *
maxtime_reader (char *lp)
{
  char *errstr;
  unsigned int max_frames;
  arg_type args[] =
  {
    { "time", uns_t, optional },
    { "frames", uns_t, optional },
    { NULL }
  };

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);

  get_arg ("time", maxtime.time);
  get_arg ("frames", max_frames);

  if (maxtime.time == 0)
    maxtime.time = max_frames * frametime;
  else if (max_frames > 0)
    return_error ("only one of frames or time is allowed");

  if (maxtime.time == 0)
    return_error ("one of frames or time must be positive");
  if (maxtime.time - warm_up_time > histlen*frametime)
    input_warning ("histlen is smaller than the simulation length");
  if (maxtime.time <= warm_up_time)
    input_warning ("warmup is not smaller than the simulation length");

  return (&maxtime_stopper);
} /* maxtime_reader () */
